import dataclasses
import functools
from collections.abc import Iterator
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar

from tmt.container import container
from tmt.log import Logger

if TYPE_CHECKING:
    from tmt._compat.typing import Self
    from tmt.steps.provision import Guest


TaskResultT = TypeVar('TaskResultT')


@container
class Task(Generic[TaskResultT]):
    """
    A base class for queueable actions.

    The class provides not just the tracking, but the implementation of the said
    action as well. Child classes must implement their action functionality in
    :py:meth:`go`.

    .. note::

        This class and its subclasses must provide their own ``__init__``
        methods, and cannot rely on :py:mod:`dataclasses` generating one. This
        is caused by subclasses often adding fields without default values,
        and ``dataclasses`` module does not allow non-default fields to be
        specified after those with default values. Therefore initialize fields
        to their defaults "manually".
    """

    #: A logger to use for logging events related to the outcome.
    logger: Logger

    result: Optional[TaskResultT]

    #: Guest on which the phase was executed. May be unset, some tasks
    #: may handle multiguest actions on their own.
    guest: Optional['Guest']

    #: If set, an exception was raised by the running task, and the exception
    #: is saved in this field.
    exc: Optional[Exception]

    #: If set, the task raised :py:class:`SystemExit` exception, and wants to
    #: terminate the run completely.
    requested_exit: Optional[SystemExit]

    # Custom yet trivial `__init__` is necessary, see note in `tmt.queue.Task`.
    def __init__(self, logger: Logger, **kwargs: Any) -> None:
        self.logger = logger

        self.result = kwargs.get('result')
        self.guest = kwargs.get('guest')
        self.exc = kwargs.get('exc')
        self.requested_exit = kwargs.get('requested_exit')

    @property
    def name(self) -> str:
        """
        A name of this task.

        Left for child classes to implement, because the name depends on the
        actual task.
        """

        raise NotImplementedError

    def go(self) -> Iterator['Self']:
        """
        Perform the task.

        Called by :py:class:`Queue` machinery to accomplish the task.

        :yields: instances of the same class, describing invocations of the
            task and their outcome. The task might be executed multiple times,
            depending on how exactly it was queued, and method would yield
            corresponding results.
        """

        raise NotImplementedError


TaskT = TypeVar('TaskT', bound='Task')  # type: ignore[type-arg]


def prepare_loggers(logger: Logger, labels: list[str]) -> dict[str, Logger]:
    """
    Create loggers for a set of labels.

    Guests are assumed to be a group a phase would be executed on, and
    therefore their labels need to be set, to provide context, plus their
    labels need to be properly aligned for more readable output.
    """

    loggers: dict[str, Logger] = {}

    # First, spawn all loggers, and set their labels if needed.
    # Don't bother with labels if there's just a single guest.
    for label in labels:
        new_logger = logger.clone()

        if len(labels) > 1:
            new_logger.labels.append(label)

        loggers[label] = new_logger

    # Second, find the longest label, and instruct all loggers to pad their
    # labels to match this length. This should create well-indented messages.
    max_label_span = max(new_logger.labels_span for new_logger in loggers.values())

    for new_logger in loggers.values():
        new_logger.labels_padding = max_label_span

    return loggers


@container
class GuestlessTask(Task[TaskResultT]):
    """
    A task not assigned to a particular set of guests.

    An extension of the :py:class:`Task` class, provides a quite generic wrapper
    for the actual task which takes care of catching exceptions and proper
    reporting.
    """

    # Custom yet trivial `__init__` is necessary, see note in `tmt.queue.Task`.
    def __init__(self, logger: Logger, **kwargs: Any) -> None:
        super().__init__(logger, **kwargs)

    def run(self, logger: Logger) -> TaskResultT:
        """
        Perform the task.

        Called once from :py:meth:`go`. Subclasses of :py:class:`GuestlessTask`
        should implement their logic in this method rather than in
        :py:meth:`go` which is already provided. If your task requires different
        handling in :py:class:`go`, it might be better derived directly from
        :py:class:`Task`.
        """

        raise NotImplementedError

    def go(self) -> Iterator['Self']:
        """
        Perform the task.

        Called by :py:class:`Queue` machinery to accomplish the task. It expects
        the child class would implement :py:meth:`run`, with ``go`` taking care
        of task/queue interaction.

        :yields: since the task is not expected to run on multiple guests,
            only a single instance of the class is yielded to describe the task
            and its outcome.
        """

        try:
            self.result = self.run(self.logger)

        except Exception as exc:
            self.result = None
            self.exc = exc

            yield self

        else:
            self.exc = None

            yield self


@container
class MultiGuestTask(Task[TaskResultT]):
    """
    A task assigned to a particular set of guests.

    An extension of the :py:class:`Task` class, provides a quite generic wrapper
    for the actual task which takes care of catching exceptions and proper
    reporting.
    """

    guests: list['Guest']

    # Custom yet trivial `__init__` is necessary, see note in `tmt.queue.Task`.
    def __init__(self, logger: Logger, guests: list['Guest'], **kwargs: Any) -> None:
        super().__init__(logger, **kwargs)

        self.guests = guests

    @functools.cached_property
    def guest_ids(self) -> list[str]:
        return sorted([guest.multihost_name for guest in self.guests])

    def run_on_guest(self, guest: 'Guest', logger: Logger) -> None:
        """
        Perform the task.

        Called from :py:meth:`go` once for every guest to run on. Subclasses of
        :py:class:`GuestlessTask` should implement their logic in this method
        rather than in :py:meth:`go` which is already provided. If your task
        requires different handling in :py:class:`go`, it might be better
        derived directly from :py:class:`Task`.
        """

        raise NotImplementedError

    def go(self) -> Iterator['Self']:
        """
        Perform the task.

        Called by :py:class:`Queue` machinery to accomplish the task. It expects
        the child class would implement :py:meth:`run`, with ``go`` taking care
        of task/queue interaction.

        :yields: instances of the same class, describing invocations of the
            task and their outcome. For each guest, one instance would be
            yielded.
        """

        multiple_guests = len(self.guests) > 1

        new_loggers = prepare_loggers(self.logger, [guest.multihost_name for guest in self.guests])
        old_loggers: dict[str, Logger] = {}

        with ThreadPoolExecutor(max_workers=len(self.guests)) as executor:
            futures: dict[Future[None], Guest] = {}

            for guest in self.guests:
                # Swap guest's logger for the one we prepared, with labels
                # and stuff.
                #
                # We can't do the same for phases - phase is shared among
                # guests, its `self.$loggingmethod()` calls need to be
                # fixed to use a logger we pass to it through the executor.
                #
                # Possibly, the same thing should happen to guest methods as
                # well, then the phase would pass the given logger to guest
                # methods when it calls them, propagating the single logger we
                # prepared...
                old_loggers[guest.multihost_name] = guest._logger
                new_logger = new_loggers[guest.multihost_name]

                guest.inject_logger(new_logger)

                if multiple_guests:
                    new_logger.info('started', color='cyan')

                # Submit each task/guest combination (save the guest & logger
                # for later)...
                futures[executor.submit(self.run_on_guest, guest, new_logger)] = guest

            # ... and then sit and wait as they get delivered to us as they
            # finish. Unpack the guest and logger, so we could preserve logging
            # and prepare the right outcome package.
            for future in as_completed(futures):
                guest = futures[future]

                old_logger = old_loggers[guest.multihost_name]
                new_logger = new_loggers[guest.multihost_name]

                if multiple_guests:
                    new_logger.info('finished', color='cyan')

                # `Future.result()` will either 1. reraise an exception the
                # callable raised, if any, or 2. return whatever the callable
                # returned - which is `None` in our case, therefore we can
                # ignore the return value.
                try:
                    result = future.result()

                except SystemExit as exc:
                    task = dataclasses.replace(self, result=None, exc=None, requested_exit=exc)

                except Exception as exc:
                    task = dataclasses.replace(self, result=None, exc=exc, requested_exit=None)

                else:
                    task = dataclasses.replace(self, result=result, exc=None, requested_exit=None)

                task.guest = guest

                yield task

                # Don't forget to restore the original logger.
                guest.inject_logger(old_logger)


class Queue(list[TaskT]):
    """
    Queue class for running tasks
    """

    def __init__(self, name: str, logger: Logger) -> None:
        super().__init__()

        self.name = name
        self._logger = logger

    def enqueue_task(self, task: TaskT) -> None:
        """
        Put new task into a queue
        """

        self.append(task)

        self._logger.info(
            f'queued {self.name} task #{len(self)}',
            task.name,
            color='cyan',
        )

    def run(self) -> Iterator[TaskT]:
        """
        Start crunching the queued tasks.

        Tasks are executed in the order, for each task/guest
        combination a :py:class:`Task` instance is yielded.
        """

        for i, task in enumerate(self):
            self._logger.info('')

            self._logger.info(
                f'{self.name} task #{i + 1}',
                task.name,
                color='cyan',
            )

            failed_tasks: list[TaskT] = []

            for outcome in task.go():
                if outcome.exc:
                    failed_tasks.append(outcome)

                yield outcome

            # TODO: make this optional
            if failed_tasks:
                return



"""
Handle libraries
"""

from typing import TYPE_CHECKING, Optional, Union

import fmf
import fmf.utils

import tmt
import tmt.log
import tmt.utils
from tmt.base import Dependency, DependencyFile, DependencyFmfId, DependencySimple
from tmt.utils import Path

if TYPE_CHECKING:
    from tmt.libraries.beakerlib import BeakerLib
    from tmt.libraries.file import File

# A beakerlib identifier type, can be a string or a fmf id (with extra beakerlib keys)
ImportedIdentifiersType = Optional[list[Dependency]]

# A Library type, can be Beakerlib or File
LibraryType = Union['BeakerLib', 'File']

# A type for Beakerlib dependencies
LibraryDependenciesType = tuple[list[Dependency], list[Dependency], list['LibraryType']]


class LibraryError(Exception):
    """
    Used when library cannot be parsed from the identifier
    """


class Library:
    """
    General library class

    Used as parent for specific libraries like beakerlib and file
    """

    def __init__(
        self,
        *,
        parent: Optional[tmt.utils.Common] = None,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Process the library identifier and fetch the library
        """

        # Use an empty common class if parent not provided (for logging, cache)
        self.parent = parent or tmt.utils.Common(logger=logger, workdir=True)
        self._logger: tmt.log.Logger = logger

        self.identifier: Dependency
        self.format: str
        self.repo: Path
        self.name: str

    @property
    def hostname(self) -> str:
        """
        Get hostname from url or default to local
        """

        return 'local'

    @property
    def fmf_node_path(self) -> Path:
        """
        Path to fmf node
        """

        return Path(self.name.strip('/'))

    def __str__(self) -> str:
        """
        Use repo/name for string representation
        """

        return f"{self.repo}{self.name}"


def library_factory(
    *,
    identifier: Dependency,
    parent: Optional[tmt.utils.Common] = None,
    logger: tmt.log.Logger,
    source_location: Optional[Path] = None,
    target_location: Optional[Path] = None,
) -> LibraryType:
    """
    Factory function to get correct library instance
    """

    from .beakerlib import BeakerLib
    from .file import File

    if isinstance(identifier, (DependencySimple, DependencyFmfId)):
        library: LibraryType = BeakerLib(identifier=identifier, parent=parent, logger=logger)

    # File import
    #
    # ignore[reportUnnecessaryIsInstance]: pyright is correct, the test is not
    # needed given the fact `Dependency` is a union of three types, and two were
    # ruled out above. But we would like to check possible violations in runtime,
    # therefore an `else` with an exception.
    # ignore[unused-ignore]: silencing mypy's complaint about silencing
    # pyright's warning :)
    elif isinstance(identifier, DependencyFile):  # type: ignore[reportUnnecessaryIsInstance,unused-ignore]
        assert source_location is not None
        assert target_location is not None  # narrow type
        library = File(
            identifier=identifier,
            parent=parent,
            logger=logger,
            source_location=source_location,
            target_location=target_location,
        )

    # Something weird
    else:
        raise LibraryError

    # Fetch the library
    try:
        library.fetch()
    except fmf.utils.RootError as exc:
        if isinstance(library, BeakerLib):
            raise tmt.utils.SpecificationError(
                f"Repository '{library.url}' does not contain fmf metadata."
            ) from exc
        raise exc

    return library


def dependencies(
    *,
    original_require: list[Dependency],
    original_recommend: Optional[list[Dependency]] = None,
    parent: Optional[tmt.utils.Common] = None,
    imported_lib_ids: ImportedIdentifiersType = None,
    logger: tmt.log.Logger,
    source_location: Optional[Path] = None,
    target_location: Optional[Path] = None,
) -> LibraryDependenciesType:
    """
    Check dependencies for possible beakerlib libraries

    Fetch all identified libraries, check their required and recommended
    packages. Return tuple (requires, recommends, libraries) containing
    list of regular rpm package names aggregated from all fetched
    libraries, list of aggregated recommended packages and a list of
    gathered libraries (instances of the Library class).

    Avoid infinite recursion by keeping track of imported library identifiers
    and not trying to fetch those again.
    """

    # Initialize lists, use set for require & recommend
    processed_require: set[Dependency] = set()
    processed_recommend: set[Dependency] = set()
    imported_lib_ids = imported_lib_ids or []
    gathered_libraries: list[LibraryType] = []
    original_require = original_require or []
    original_recommend = original_recommend or []

    # Cut circular dependencies to avoid infinite recursion
    def already_fetched(lib: Dependency) -> bool:
        if not imported_lib_ids:
            return True
        return lib not in imported_lib_ids

    to_fetch = original_require + original_recommend
    for dependency in filter(already_fetched, to_fetch):
        # Library require/recommend
        try:
            library = library_factory(
                logger=logger,
                identifier=dependency,
                parent=parent,
                source_location=source_location,
                target_location=target_location,
            )
            gathered_libraries.append(library)
            imported_lib_ids.append(library.identifier)

            from .beakerlib import BeakerLib

            if isinstance(library, BeakerLib):
                # Recursively check for possible dependent libraries
                assert parent is not None  # narrow type
                assert parent.workdir is not None  # narrow type
                # TODO: make this one go away once fmf is properly annotated.
                # pyright detects the type might be `Unknown` because of how
                # fmf handles some corner cases, therefore `is not None` is
                # not strong enough.
                assert isinstance(library.tree.root, str)  # narrow type
                requires, recommends, libraries = dependencies(
                    original_require=library.require,
                    original_recommend=library.recommend,
                    parent=parent,
                    imported_lib_ids=imported_lib_ids,
                    logger=logger,
                    source_location=library.source_directory,
                    target_location=Path(library.tree.root),
                )
                processed_require.update(set(requires))
                processed_recommend.update(set(recommends))
                gathered_libraries.extend(libraries)
        # Regular package require/recommend
        except LibraryError:
            if dependency in original_require:
                processed_require.add(dependency)
            if dependency in original_recommend:
                processed_recommend.add(dependency)

    # Convert to list and return the results
    return list(processed_require), list(processed_recommend), gathered_libraries



import re
import shutil
from collections.abc import Iterator
from typing import Any, Literal, Optional, TypeVar, Union, cast

import fmf
import fmf.utils

import tmt
import tmt.base
import tmt.log
import tmt.options
import tmt.package_managers
import tmt.steps
import tmt.steps.prepare
import tmt.utils
from tmt.container import container, field
from tmt.package_managers import (
    FileSystemPath,
    Installable,
    Options,
    Package,
    PackagePath,
    PackageUrl,
)
from tmt.package_managers.bootc import Bootc, BootcEngine
from tmt.result import PhaseResult
from tmt.steps.provision import Guest
from tmt.utils import Command, Path, ShellScript

COPR_URL = 'https://copr.fedorainfracloud.org/coprs'


T = TypeVar('T')


class InstallBase(tmt.utils.Common):
    """
    Base class for installation implementations
    """

    guest: Guest

    skip_missing: bool = False
    exclude: list[Package]

    packages: list[Union[Package, FileSystemPath]]
    local_packages: list[PackagePath]
    remote_packages: list[PackageUrl]
    debuginfo_packages: list[Package]

    package_directory: Path

    def __init__(
        self,
        *,
        parent: 'PrepareInstall',
        guest: Guest,
        dependencies: list[tmt.base.DependencySimple],
        directories: list[Path],
        exclude: list[str],
        logger: tmt.log.Logger,
        **kwargs: Any,
    ) -> None:
        """
        Initialize installation data
        """

        super().__init__(logger=logger, parent=parent, relative_indent=0, guest=guest, **kwargs)

        if not dependencies and not directories:
            self.debug("No packages for installation found.", level=3)

        self.guest = guest
        self.exclude = [Package(package) for package in exclude]

        self.skip_missing = bool(parent.get('missing') == 'skip')

        # Prepare package lists and installation command
        self.prepare_installables(dependencies, directories)

    def prepare_installables(
        self,
        dependencies: list[tmt.base.DependencySimple],
        directories: list[Path],
    ) -> None:
        """
        Process package names and directories
        """

        self.packages = []
        self.local_packages = []
        self.remote_packages = []
        self.debuginfo_packages = []

        # Detect local, debuginfo and repository packages
        for dependency in dependencies:
            if re.match(r"^http(s)?://", dependency):
                self.remote_packages.append(PackageUrl(dependency))
            elif dependency.endswith(".rpm"):
                self.local_packages.append(PackagePath(dependency))
            elif re.search(r"-debug(info|source)(\.|$)", dependency):
                # Strip the '-debuginfo' string from package name
                # (installing with it doesn't work on RHEL7)
                self.debuginfo_packages.append(
                    Package(re.sub(r"-debuginfo((?=\.)|$)", "", str(dependency)))
                )

            elif dependency.startswith('/'):
                self.packages.append(FileSystemPath(dependency))

            else:
                self.packages.append(Package(dependency))

        # Check rpm packages in local directories
        for directory in directories:
            self.info('directory', directory, 'green')
            if not directory.is_dir():
                raise tmt.utils.PrepareError(f"Packages directory '{directory}' not found.")
            for filepath in directory.iterdir():
                if filepath.suffix == '.rpm':
                    self.debug(f"Found rpm '{filepath}'.", level=3)
                    self.local_packages.append(PackagePath(filepath))

    def prepare_repository(self) -> None:
        """
        Configure additional repository
        """

        raise NotImplementedError

    def list_installables(self, title: str, *installables: Installable) -> Iterator[Installable]:
        """
        Show package info and return package names
        """

        # Show a brief summary by default
        if not self.verbosity_level:
            summary = fmf.utils.listed(installables, max=3)
            self.info(title, summary, 'green')
        # Provide a full list of packages in verbose mode
        else:
            summary = fmf.utils.listed(installables, 'package')
            self.info(title, summary + ' requested', 'green')
            for package in sorted(installables):
                self.verbose(str(package), shift=1)

        yield from installables

    def prepare_install_local(self) -> None:
        """
        Copy packages to the test system
        """

        assert self.parent is not None
        # FIXME: cast() - https://github.com/teemtee/tmt/issues/1372
        workdir = cast(PrepareInstall, self.parent).step.workdir
        if not workdir:
            raise tmt.utils.GeneralError('workdir should not be empty')
        self.package_directory = workdir / 'packages'
        self.package_directory.mkdir(parents=True)

        # Copy local packages into workdir, push to guests
        for package in self.local_packages:
            self.verbose(package.name, shift=1)
            self.debug(f"Copy '{package}' to '{self.package_directory}'.", level=3)
            shutil.copy(package, self.package_directory)
        self.guest.push()

    def install_from_repository(self) -> None:
        """
        Default base install method for packages from repositories
        """

    def install_local(self) -> None:
        """
        Default base install method for local packages
        """

    def install_from_url(self) -> None:
        """
        Default base install method for packages which are from URL
        """

    def install_debuginfo(self) -> None:
        """
        Default base install method for debuginfo packages
        """

    def install(self) -> None:
        """
        Perform the actual installation
        """

        try:
            self._install()

        except Exception as exc1:
            # We do not have any special handling for exceptions raised by the following code.
            # Wrapping them with try/except gives us a chance to attach the original exception
            # to whatever the code may raise, and therefore preserve the information attached
            # to the original exception.
            self.warn('Installation failed, trying again after metadata refresh.')

            try:
                # Refresh cache in case of recent but not updated change do repodata
                self.guest.package_manager.refresh_metadata()
                self._install()

            except Exception as exc2:
                raise exc2 from exc1

    def _install(self) -> None:
        """
        Helper method to perform the actual installation steps
        """

        if self.local_packages:
            self.prepare_install_local()
            self.install_local()
        if self.remote_packages:
            self.install_from_url()
        if self.packages:
            self.install_from_repository()
        if self.debuginfo_packages:
            self.install_debuginfo()

    def rpm_check(self, package: str, mode: str = '-q') -> None:
        """
        Run rpm command to check package existence

        Throws tmt.utils.RunError
        """

        output = self.guest.execute(Command('rpm', mode, package), silent=True)
        assert output.stdout
        self.debug(f"Package '{output.stdout.strip()}' already installed.")


class Copr(tmt.utils.Common):
    copr_plugin: str

    guest: Guest

    # Keep this method around, to correctly support Python's method resolution order.
    def __init__(self, *args: Any, guest: Guest, **kwargs: Any) -> None:
        super().__init__(*args, guest=guest, **kwargs)

        self.guest = guest

    def enable_copr_epel6(self, copr: str) -> None:
        """
        Manually enable copr repositories for epel6
        """

        # Parse the copr repo name
        matched = re.match("^(@)?([^/]+)/([^/]+)$", copr)
        if not matched:
            raise tmt.utils.PrepareError(f"Invalid copr repository '{copr}'.")
        group, name, project = matched.groups()
        group = 'group_' if group else ''
        # Prepare the repo file url
        parts = [COPR_URL] + (['g'] if group else [])
        parts += [name, project, 'repo', 'epel-6']
        parts += [f"{group}{name}-{project}-epel-6.repo"]
        url = '/'.join(parts)
        # Download the repo file on guest
        try:
            self.guest.execute(
                Command('curl', '-LOf', url),
                cwd=Path('/etc/yum.repos.d'),
                silent=True,
            )
        except tmt.utils.RunError as error:
            if error.stderr and 'not found' in error.stderr.lower():
                raise tmt.utils.PrepareError(f"Copr repository '{copr}' not found.")
            raise

    def enable_copr(self, repositories: list[str]) -> None:
        """
        Enable requested copr repositories
        """

        if not repositories:
            return

        package_manager = self.guest.package_manager

        # Try to install copr plugin
        self.debug('Make sure the copr plugin is available.')
        try:
            package_manager.install(Package(self.copr_plugin))

        # Enable repositories manually for epel6
        except tmt.utils.RunError:
            for repository in repositories:
                self.info('copr', repository, 'green')
                self.enable_copr_epel6(repository)

        # Enable repositories using copr plugin
        else:
            for repository in repositories:
                self.info('copr', repository, 'green')

                self.guest.execute(
                    ShellScript(
                        f"{package_manager.engine.command.to_script()} copr "
                        f"{package_manager.engine.options.to_script()} enable -y {repository}"
                    )
                )


class InstallDnf(InstallBase, Copr):
    """
    Install packages using dnf
    """

    copr_plugin = "dnf-plugins-core"

    def install_local(self) -> None:
        """
        Install packages stored in a local directory
        """

        # Use both dnf install/reinstall to get all packages refreshed
        # FIXME Simplify this once BZ#1831022 is fixed/implemented.
        filelist = [
            PackagePath(self.package_directory / filename.name) for filename in self.local_packages
        ]

        self.guest.package_manager.install(
            *filelist,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
                check_first=False,
            ),
        )

        self.guest.package_manager.reinstall(
            *filelist,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
                check_first=False,
            ),
        )

        summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
        self.info('total', f"{summary} installed", 'green')

    def install_from_url(self) -> None:
        """
        Install packages stored on a remote URL
        """

        self.guest.package_manager.install(
            *self.list_installables("remote package", *self.remote_packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_from_repository(self) -> None:
        """
        Install packages from a repository
        """

        self.guest.package_manager.install(
            *self.list_installables("package", *self.packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_debuginfo(self) -> None:
        """
        Install debuginfo packages
        """

        packages = self.list_installables("debuginfo", *self.debuginfo_packages)

        self.guest.package_manager.install_debuginfo(
            *packages,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

        # Check the packages are installed on the guest because 'debuginfo-install'
        # returns 0 even though it didn't manage to install the required packages
        if not self.skip_missing:
            self.guest.package_manager.check_presence(
                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
            )


class InstallDnf5(InstallDnf):
    """
    Install packages using dnf5
    """

    copr_plugin = "dnf5-command(copr)"


class InstallYum(InstallDnf):
    """
    Install packages using yum
    """

    copr_plugin = "yum-plugin-copr"


class InstallRpmOstree(InstallBase, Copr):
    """
    Install packages using rpm-ostree
    """

    copr_plugin = "dnf-plugins-core"

    recommended_packages: list[Union[Package, FileSystemPath]]
    required_packages: list[Union[Package, FileSystemPath]]

    def enable_copr(self, repositories: list[str]) -> None:
        """
        Enable requested copr repositories
        """

        # rpm-ostree can step outside of its zone of competence, and use
        # another package manager to enable copr. We just need to cheat
        # the `InstallDnf5` & swap guest's package manager for `dnf5`
        # for a moment.
        self.guest.facts.package_manager = 'dnf5'
        # TODO: wouldn't it be nice if we could just call copr method
        # without populating `dependencies` & co. with dummy values.
        InstallDnf5(
            parent=cast('PrepareInstall', self.parent),
            guest=self.guest,
            dependencies=[],
            directories=[],
            exclude=[],
            logger=self._logger,
        ).enable_copr(repositories)
        self.guest.facts.package_manager = 'rpm-ostree'

    def sort_packages(self) -> None:
        """
        Identify required and recommended packages
        """

        self.recommended_packages = []
        self.required_packages = []
        for package in self.packages:
            presence = self.guest.package_manager.check_presence(package)

            if not all(presence.values()):
                if self.skip_missing:
                    self.recommended_packages.append(package)
                else:
                    self.required_packages.append(package)

    def install_debuginfo(self) -> None:
        """
        Install debuginfo packages
        """

        self.warn("Installation of debuginfo packages not supported yet.")

    def install_local(self) -> None:
        """
        Install copied local packages
        """

        local_packages_installed: list[PackagePath] = []
        for package in self.local_packages:
            try:
                self.guest.package_manager.install(
                    PackagePath(self.package_directory / package.name),
                    options=Options(check_first=False),
                )
                local_packages_installed.append(package)
            except tmt.utils.RunError as error:
                self.warn(f"Local package '{package}' not installed: {error.stderr}")
        summary = fmf.utils.listed(local_packages_installed, 'local package')
        self.info('total', f"{summary} installed", 'green')

    def install_from_repository(self) -> None:
        """
        Install packages from the repository
        """

        self.sort_packages()

        # Install recommended packages
        if self.recommended_packages:
            self.list_installables("package", *self.recommended_packages)
            for package in self.recommended_packages:
                try:
                    self.guest.package_manager.install(package)
                except tmt.utils.RunError as error:
                    self.debug(f"Package installation failed: {error}")
                    self.warn(f"Unable to install recommended package '{package}'.")
                    continue

        # Install required packages
        if self.required_packages:
            self.guest.package_manager.install(
                *self.list_installables("package", *self.required_packages)
            )


class InstallApt(InstallBase):
    """
    Install packages using apt
    """

    def install_local(self) -> None:
        """
        Install packages stored in a local directory
        """

        filelist = [
            PackagePath(self.package_directory / filename) for filename in self.local_packages
        ]

        self.guest.package_manager.install(
            *self.list_installables('local packages', *filelist),
            options=Options(
                excluded_packages=self.exclude, skip_missing=self.skip_missing, check_first=False
            ),
        )

        summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
        self.info('total', f"{summary} installed", 'green')

    def install_from_url(self) -> None:
        """
        Install packages stored on a remote URL
        """

        self.guest.package_manager.install(
            *self.list_installables("remote package", *self.remote_packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_from_repository(self) -> None:
        """
        Install packages from a repository
        """

        self.guest.package_manager.install(
            *self.list_installables("package", *self.packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_debuginfo(self) -> None:
        """
        Install debuginfo packages
        """

        packages = self.list_installables("debuginfo", *self.debuginfo_packages)

        self.guest.package_manager.install_debuginfo(
            *packages,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

        # Check the packages are installed on the guest because 'debuginfo-install'
        # returns 0 even though it didn't manage to install the required packages
        if not self.skip_missing:
            self.guest.package_manager.check_presence(
                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
            )


class InstallBootc(InstallBase):
    """Install packages using bootc container image mode"""

    @property
    def _engine(self) -> BootcEngine:
        assert isinstance(self.guest.package_manager.engine, BootcEngine)

        return self.guest.package_manager.engine

    def install_from_repository(self) -> None:
        self._engine.install(
            *self.list_installables("package", *self.packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_from_url(self) -> None:
        self._engine.install(
            *self.list_installables("remote package", *self.remote_packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_local(self) -> None:
        # Make sure the containerfile session has been initialized. The
        # engine would do it for us, but we need to prepend some
        # directives first before using the engine.
        self._engine.open_containerfile_directives()

        # Filelist for packages on the guest
        filelist = [
            PackagePath(self.package_directory / filename.name) for filename in self.local_packages
        ]

        self._engine.containerfile_directives.append(f'RUN mkdir -p {self.package_directory}')
        self._engine.containerfile_directives.append(
            f'COPY {" ".join(str(file) for file in filelist)} {self.package_directory}'
        )

        self._engine.install(
            *filelist,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
                check_first=False,
            ),
        )

        self._engine.reinstall(
            *filelist,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
                check_first=False,
            ),
        )

    def install_debuginfo(self) -> None:
        packages = self.list_installables("debuginfo", *self.debuginfo_packages)

        self._engine.install_debuginfo(
            *packages,
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

        # Check the packages are installed on the guest because 'debuginfo-install'
        # returns 0 even though it didn't manage to install the required packages
        if not self.skip_missing:
            self._engine.check_presence(
                *[Package(f'{package}-debuginfo') for package in self.debuginfo_packages]
            )

    def _install(self) -> None:
        """Coordinate installation process through containerfile building and switching"""
        # Call base install methods to collect all package types
        super()._install()

        cast(Bootc, self.guest.package_manager).build_container()


class InstallApk(InstallBase):
    """
    Install packages using apk
    """

    def install_local(self) -> None:
        """
        Install packages stored in a local directory
        """

        filelist = [
            PackagePath(self.package_directory / filename) for filename in self.local_packages
        ]

        self.guest.package_manager.install(
            *self.list_installables('local packages', *filelist),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
                allow_untrusted=True,
                check_first=False,
            ),
        )

        summary = fmf.utils.listed([str(path) for path in self.local_packages], 'local package')
        self.info('total', f"{summary} installed", 'green')

    def install_from_url(self) -> None:
        """
        Install packages stored on a remote URL
        """

        raise tmt.utils.PrepareError(
            f'Package manager "{self.guest.facts.package_manager}" '
            'does not support installing from a remote URL.'
        )

    def install_from_repository(self) -> None:
        """
        Install packages from a repository
        """

        self.guest.package_manager.install(
            *self.list_installables("package", *self.packages),
            options=Options(
                excluded_packages=self.exclude,
                skip_missing=self.skip_missing,
            ),
        )

    def install_debuginfo(self) -> None:
        """
        Install debuginfo packages
        """

        raise tmt.utils.PrepareError(
            f'Package manager "{self.guest.facts.package_manager}" does not support '
            'installing debuginfo packages.'
        )


@container
class PrepareInstallData(tmt.steps.prepare.PrepareStepData):
    package: list[tmt.base.DependencySimple] = field(
        default_factory=list,
        option=('-p', '--package'),
        metavar='PACKAGE',
        multiple=True,
        help='Package name or path to rpm to be installed.',
        # PrepareInstall supports *simple* requirements only
        normalize=lambda key_address, value, logger: tmt.base.assert_simple_dependencies(
            tmt.base.normalize_require(key_address, value, logger),
            "'install' plugin support simple packages only, no fmf links are allowed",
            logger,
        ),
        serialize=lambda packages: [package.to_spec() for package in packages],
        unserialize=lambda serialized: [
            tmt.base.DependencySimple.from_spec(package) for package in serialized
        ],
    )

    directory: list[Path] = field(
        default_factory=list,
        option=('-D', '--directory'),
        metavar='PATH',
        multiple=True,
        help='Path to a local directory with rpm packages.',
        normalize=tmt.utils.normalize_path_list,
    )

    copr: list[str] = field(
        default_factory=list,
        option=('-c', '--copr'),
        metavar='REPO',
        multiple=True,
        help='Copr repository to be enabled.',
        normalize=tmt.utils.normalize_string_list,
    )

    exclude: list[str] = field(
        default_factory=list,
        option=('-x', '--exclude'),
        metavar='PACKAGE',
        multiple=True,
        help='Packages to be skipped during installation.',
        normalize=tmt.utils.normalize_string_list,
    )

    # TODO: use enum
    missing: Literal['skip', 'fail'] = field(
        default='fail',
        option=('-m', '--missing'),
        metavar='ACTION',
        choices=['fail', 'skip'],
        help='Action on missing packages, fail (default) or skip.',
    )


@tmt.steps.provides_method('install')
class PrepareInstall(tmt.steps.prepare.PreparePlugin[PrepareInstallData]):
    """
    Install packages on the guest.

    Example config:

    .. code-block:: yaml

        prepare:
            how: install
            copr: psss/tmt
            package: tmt-all
            missing: fail

    Use ``copr`` for enabling a desired Copr repository and ``missing`` to choose
    whether missing packages should be silently ignored (``skip``) or a
    preparation error should be reported (``fail``), which is the default.

    One or more RPM packages can be specified under the
    ``package`` attribute. The packages will be installed
    on the guest. They can either be specified using their
    names, paths to local rpm files or urls to remote rpms.

    .. code-block:: yaml

        # Install local rpms using file path
        prepare:
            how: install
            package:
                - tmp/RPMS/noarch/tmt-0.15-1.fc31.noarch.rpm
                - tmp/RPMS/noarch/python3-tmt-0.15-1.fc31.noarch.rpm

    .. code-block:: yaml

        # Install remote packages using url
        prepare:
            how: install
            package:
              - https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm
              - https://dl.fedoraproject.org/pub/epel/epel-next-release-latest-8.noarch.rpm

    .. code-block:: yaml

        # Install the whole directory, exclude selected packages
        prepare:
            how: install
            directory:
              - tmp/RPMS/noarch
            exclude:
              - tmt+all
              - tmt+provision-virtual

    .. code-block:: yaml

        prepare:
            how: install
            # Repository with a group owner (@ prefixed) requires quotes, e.g.
            # copr: "@osci/rpminspect"
            copr: psss/tmt
            package: tmt-all
            missing: skip

    Use ``directory`` to install all packages from given folder and
    ``exclude`` to skip selected packages (globbing characters are supported as
    well).

    .. code-block:: yaml

        prepare:
            how: install
            directory: tmp/RPMS/noarch
            exclude: tmt+provision-virtual

    .. note::

        When testing ostree booted deployments tmt will use
        ``rpm-ostree`` as the package manager to perform the installation of
        requested packages. The current limitations of the ``rpm-ostree``
        implementation are:

        * Cannot install new version of already installed local rpm.
        * No support for installing debuginfo packages at this time.
    """

    _data_class = PrepareInstallData

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Perform preparation for the guests
        """

        results = super().go(guest=guest, environment=environment, logger=logger)

        # Nothing to do in dry mode
        if self.is_dry_run:
            return results

        # Pick the right implementation
        # TODO: it'd be nice to use a "plugin registry" and make the
        # implementations discovered as any other plugins. Package managers are
        # shipped as plugins, but we still need a matching *installation* class.
        # But do we really need a class per package manager family? Maybe the
        # code could be integrated into package manager plugins directly.
        if guest.facts.package_manager == 'bootc':
            installer: InstallBase = InstallBootc(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'rpm-ostree':
            installer = InstallRpmOstree(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'dnf5':
            installer = InstallDnf5(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'dnf':
            installer = InstallDnf(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'yum':
            installer = InstallYum(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'apt':
            installer = InstallApt(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager == 'apk':
            installer = InstallApk(
                logger=logger,
                parent=self,
                dependencies=self.data.package,
                directories=self.data.directory,
                exclude=self.data.exclude,
                guest=guest,
            )

        elif guest.facts.package_manager is None:
            raise tmt.utils.PrepareError('Unrecognized package manager.')

        else:
            raise tmt.utils.PrepareError(
                f"Package manager '{guest.facts.package_manager}' "
                "is not supported by 'prepare/install'."
            )

        # Enable copr repositories...
        if isinstance(installer, Copr):
            installer.enable_copr(self.data.copr)

        # ... and install packages.
        installer.install()

        return results



#!/usr/bin/python3
"""
A simple script to parse the output of 'tmt test export .'
"""

import sys

import ruamel.yaml


def verify(obj, obj_id, field, exp):
    got = obj[field]
    sys.stdout.write(
        f">>> Verify {obj_id}'s field '{field}' ......\n"
        f"...\texp: {exp}\n"
        f"...\tgot: {got}\n"
        f"...\tres: "
    )
    assert got == exp
    sys.stdout.write("PASSED\n\n")


def main(argc, argv):
    yaml = ruamel.yaml.YAML()
    tests = yaml.load(sys.stdin)

    test01 = tests[0]
    verify(test01, 'test01', 'name', '/tests/enabled/default')
    verify(test01, 'test01', 'summary', 'This test is enabled by default')
    verify(test01, 'test01', 'description', None)
    verify(test01, 'test01', 'contact', [])
    verify(test01, 'test01', 'component', [])
    verify(test01, 'test01', 'test', 'true')
    verify(test01, 'test01', 'path', '/')
    verify(test01, 'test01', 'framework', 'shell')
    verify(test01, 'test01', 'manual', False)
    verify(test01, 'test01', 'require', [])
    verify(test01, 'test01', 'recommend', [])
    verify(test01, 'test01', 'environment', {})
    verify(test01, 'test01', 'duration', '5m')
    verify(test01, 'test01', 'enabled', True)
    verify(test01, 'test01', 'order', 50)
    verify(test01, 'test01', 'result', 'respect')
    verify(test01, 'test01', 'tag', [])
    verify(test01, 'test01', 'tier', None)
    verify(test01, 'test01', 'link', [])

    test02 = tests[1]
    verify(test02, 'test02', 'name', '/tests/enabled/disabled')
    verify(test02, 'test02', 'summary', 'This test is disabled')
    verify(test02, 'test02', 'description', None)
    verify(test02, 'test02', 'contact', [])
    verify(test02, 'test02', 'component', [])
    verify(test02, 'test02', 'test', 'true')
    verify(test02, 'test02', 'path', '/')
    verify(test02, 'test02', 'framework', 'shell')
    verify(test02, 'test02', 'manual', False)
    verify(test02, 'test02', 'require', [])
    verify(test02, 'test02', 'recommend', [])
    verify(test02, 'test02', 'environment', {})
    verify(test02, 'test02', 'duration', '5m')
    verify(test02, 'test02', 'enabled', False)
    verify(test02, 'test02', 'order', 50)
    verify(test02, 'test02', 'result', 'respect')
    verify(test02, 'test02', 'tag', [])
    verify(test02, 'test02', 'tier', None)
    verify(test02, 'test02', 'link', [])
    return 0


if __name__ == '__main__':
    sys.exit(main(len(sys.argv), sys.argv))



import copy
from typing import TYPE_CHECKING, Any, Literal, Optional, TypeVar, Union, cast

import click
import fmf
import fmf.utils

import tmt
import tmt.log
import tmt.steps
import tmt.steps.discover
import tmt.steps.provision
import tmt.utils
from tmt.container import container, simple_field
from tmt.options import option
from tmt.plugins import PluginRegistry
from tmt.result import PhaseResult
from tmt.steps import (
    Action,
    ActionTask,
    PhaseQueue,
    PluginTask,
    PullTask,
    PushTask,
    sync_with_guests,
)
from tmt.steps.provision import Guest
from tmt.utils import uniq

if TYPE_CHECKING:
    import tmt.base
    import tmt.cli
    from tmt.base import Plan


@container
class PrepareStepData(tmt.steps.WhereableStepData, tmt.steps.StepData):
    pass


PrepareStepDataT = TypeVar('PrepareStepDataT', bound=PrepareStepData)


class _RawPrepareStepData(tmt.steps._RawStepData, tmt.steps.RawWhereableStepData, total=False):
    pass


class PreparePlugin(tmt.steps.Plugin[PrepareStepDataT, list[PhaseResult]]):
    """
    Common parent of prepare plugins
    """

    # ignore[assignment]: as a base class, PrepareStepData is not included in
    # PrepareStepDataT.
    _data_class = PrepareStepData  # type: ignore[assignment]

    # Methods ("how: ..." implementations) registered for the same step.
    _supported_methods: PluginRegistry[tmt.steps.Method] = PluginRegistry('step.prepare')

    @classmethod
    def base_command(
        cls,
        usage: str,
        method_class: Optional[type[click.Command]] = None,
    ) -> click.Command:
        """
        Create base click command (common for all prepare plugins)
        """

        # Prepare general usage message for the step
        if method_class:
            usage = Prepare.usage(method_overview=usage)

        # Create the command
        @click.command(cls=method_class, help=usage)
        @click.pass_context
        @option(
            '-h',
            '--how',
            metavar='METHOD',
            help='Use specified method for environment preparation.',
        )
        @tmt.steps.PHASE_OPTIONS
        def prepare(context: 'tmt.cli.Context', **kwargs: Any) -> None:
            context.obj.steps.add('prepare')
            Prepare.store_cli_invocation(context)

        return prepare

    def go(
        self,
        *,
        guest: 'tmt.steps.provision.Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Prepare the guest (common actions)
        """

        self.go_prolog(logger)

        # Show guest name first in multihost scenarios
        if self.step.plan.provision.is_multihost:
            logger.info('guest', guest.name, 'green')

        # Show requested role if defined
        if self.data.where:
            logger.info('where', fmf.utils.listed(self.data.where), 'green')

        return []


# Required & recommended packages
#
# Structures and code for collecting requirements for different guests
# or their groups to avoid installing all test requirements on all
# guests. For example, a test running on the "server" guest might
# require package `foo` while the test running on the "client" might
# require package `bar`, and `foo` and `bar` cannot be installed at the
# same time.


@container
class DependencyCollection:
    """
    Bundle guests and packages to install on them
    """

    # Guest*s*, not a guest. The list will start with just one guest at
    # first, but when grouping guests by same requirements, we'd start
    # adding guests to the list when spotting same set of dependencies.
    guests: list[Guest]
    dependencies: list['tmt.base.DependencySimple'] = simple_field(default_factory=list)

    @property
    def as_key(self) -> 'DependencyCollectionKey':
        return frozenset(self.dependencies)


DependencyCollectionKey = frozenset['tmt.base.DependencySimple']


class Prepare(tmt.steps.Step):
    """
    Prepare the environment for testing.

    Use the 'order' attribute to select in which order preparation
    should happen if there are multiple configs. Default order is '50'.
    Default order of required packages installation is '70', for the
    recommended packages it is '75'.
    """

    _plugin_base_class = PreparePlugin

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        return {*super()._preserved_workdir_members, 'results.yaml'}

    def __init__(
        self,
        *,
        plan: 'Plan',
        data: tmt.steps.RawStepDataArgument,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Initialize prepare step data
        """

        super().__init__(plan=plan, data=data, logger=logger)
        self.preparations_applied = 0

    def wake(self) -> None:
        """
        Wake up the step (process workdir and command line)
        """

        super().wake()

        # Choose the right plugin and wake it up
        for data in self.data:
            # FIXME: cast() - see https://github.com/teemtee/tmt/issues/1599
            plugin = cast(PreparePlugin[PrepareStepData], PreparePlugin.delegate(self, data=data))
            plugin.wake()
            # Add plugin only if there are data
            if not plugin.data.is_bare:
                self._phases.append(plugin)

        # Nothing more to do if already done and not asked to run again
        if self.status() == 'done' and not self.should_run_again:
            self.debug('Prepare wake up complete (already done before).', level=2)
        # Save status and step data (now we know what to do)
        else:
            self.status('todo')
            self.save()

    def summary(self) -> None:
        """
        Give a concise summary of the preparation
        """

        preparations = fmf.utils.listed(self.preparations_applied, 'preparation')
        self.info('summary', f'{preparations} applied', 'green', shift=1)

    def go(self, force: bool = False) -> None:
        """
        Prepare the guests
        """

        super().go(force=force)

        # Nothing more to do if already done
        if self.status() == 'done':
            self.info('status', 'done', 'green', shift=1)
            self.summary()
            self.actions()
            return

        import tmt.base

        # All phases from all steps.
        phases = [
            phase
            for step in (
                self.plan.discover,
                self.plan.provision,
                self.plan.prepare,
                self.plan.execute,
                self.plan.finish,
                self.plan.report,
            )
            for phase in step.phases(classes=step._plugin_base_class)
        ]

        # All provisioned guests.
        guests = self.plan.provision.ready_guests

        # 1. collect all requirements, per guest. For each phase, test,
        # check and so on, find out on which guest it needs to run, and
        # add its requirements to a guest-specific collection.

        # Collecting all essential requirements, per guest.
        collected_essential_requires = {
            guest: DependencyCollection(guests=[guest]) for guest in guests
        }

        # Collecting all required packages, per guest.
        collected_requires = {guest: DependencyCollection(guests=[guest]) for guest in guests}

        # Collecting all recommended packages, per guest.
        collected_recommends = {guest: DependencyCollection(guests=[guest]) for guest in guests}

        # For each guest, check everything that can run, and if enabled
        # for the given guest, add requirements into the correct
        # collection.
        for guest in guests:
            # First, check phases - plugins have their own requirements,
            # the essential requirements.
            for phase in phases:
                if not phase.enabled_by_when:
                    continue
                if not phase.enabled_on_guest(guest):
                    continue

                collected_essential_requires[
                    guest
                ].dependencies += tmt.base.assert_simple_dependencies(
                    # ignore[attr-defined]: mypy thinks that phase is Phase type, while its
                    # actually PluginClass
                    phase.essential_requires(),  # type: ignore[attr-defined]
                    'After beakerlib processing, tests may have only simple requirements',
                    self._logger,
                )

            # The `discover` step is different: no phases, just query tests
            # collected by the step itself. Maybe we could iterate over
            # `discover` phases, but I think re-runs and workdir reuse would
            # use what the step loads from its storage, `tests.yaml`. Which
            # means, there probably would be no phases to inspect from time to
            # time, therefore going after the step itself.
            for test_origin in self.plan.discover.tests(enabled=True):
                test = test_origin.test

                if not test.enabled_on_guest(guest):
                    continue

                collected_requires[guest].dependencies += tmt.base.assert_simple_dependencies(
                    test.require,
                    'After beakerlib processing, tests may have only simple requirements',
                    self._logger,
                )

                collected_recommends[guest].dependencies += tmt.base.assert_simple_dependencies(
                    test.recommend,
                    'After beakerlib processing, tests may have only simple requirements',
                    self._logger,
                )

                collected_essential_requires[
                    guest
                ].dependencies += test.test_framework.get_requirements(test, self._logger)

                for check in test.check:
                    collected_essential_requires[
                        guest
                    ].dependencies += check.plugin.essential_requires(guest, test, self._logger)

        # 2. Now we have guests and all their requirements. There can be
        # duplicities, multiple tests requesting the same package, but also
        # some guests may share the set of packages to be installed on them.
        # Let's say N guests share a `role`, all their tests would add the same
        # requirements to these guests.
        #
        # So the final 2 steps:
        #
        # 1. make the list of requirements unique,
        # 2. group guests with same requirements.
        def _prune_collections(
            collections: dict[Guest, DependencyCollection],
        ) -> list[DependencyCollection]:
            pruned: dict[DependencyCollectionKey, DependencyCollection] = {}

            for guest, collection in collections.items():
                collection.dependencies = uniq(collection.dependencies)

                if collection.as_key in pruned:
                    pruned[collection.as_key].guests.append(guest)

                else:
                    pruned[collection.as_key] = collection

            return list(pruned.values())

        pruned_essential_requires = _prune_collections(collected_essential_requires)
        pruned_requires = _prune_collections(collected_requires)
        pruned_recommends = _prune_collections(collected_recommends)

        # 3. for each collection, which now groups a set of packages and
        # all guests they need to be installed on, add new phase that
        # would take care of installation.
        def _emit_phase(
            pruned_collections: list[DependencyCollection],
            name: str,
            summary: str,
            order: int,
            missing: Literal['skip', 'fail'] = 'fail',
        ) -> None:
            from tmt.steps.prepare.install import PrepareInstallData

            for collection in pruned_collections:
                if not collection.dependencies:
                    continue

                data = PrepareInstallData(
                    name=name,
                    how='install',
                    summary=summary,
                    order=order,
                    where=[guest.name for guest in collection.guests],
                    package=collection.dependencies,
                    missing=missing,
                )

                self._phases.append(PreparePlugin.delegate(self, data=data))

        _emit_phase(
            pruned_essential_requires,
            'essential-requires',
            'Install essential required packages',
            tmt.steps.PHASE_ORDER_PREPARE_INSTALL_ESSENTIAL_REQUIRES,
        )

        _emit_phase(
            pruned_requires,
            'requires',
            'Install required packages',
            tmt.steps.PHASE_ORDER_PREPARE_INSTALL_REQUIRES,
        )

        _emit_phase(
            pruned_recommends,
            'recommends',
            'Install recommended packages',
            tmt.steps.PHASE_ORDER_PREPARE_INSTALL_RECOMMENDS,
            missing='skip',
        )

        # Prepare guests (including workdir sync)
        guest_copies: list[Guest] = []

        for guest in self.plan.provision.ready_guests:
            # Create a guest copy and change its parent so that the
            # operations inside prepare plugins on the guest use the
            # prepare step config rather than provision step config.
            guest_copy = copy.copy(guest)
            guest_copy.inject_logger(
                guest._logger.clone().apply_verbosity_options(**self._cli_options)
            )
            guest_copy.parent = self

            guest_copies.append(guest_copy)

        if guest_copies:
            sync_with_guests(
                self, 'push', PushTask(logger=self._logger, guests=guest_copies), self._logger
            )

            # To separate "push" from "prepare" queue visually
            self.info('')

        queue: PhaseQueue[PrepareStepData, list[PhaseResult]] = PhaseQueue(
            'prepare', self._logger.descend(logger_name=f'{self}.queue')
        )

        for prepare_phase in self.phases(classes=(Action, PreparePlugin)):
            if isinstance(prepare_phase, Action):
                queue.enqueue_action(phase=prepare_phase)

            elif prepare_phase.enabled_by_when:
                queue.enqueue_plugin(
                    phase=prepare_phase,  # type: ignore[arg-type]
                    guests=[
                        guest for guest in guest_copies if prepare_phase.enabled_on_guest(guest)
                    ],
                )

        failed_tasks: list[Union[ActionTask, PluginTask[PrepareStepData, list[PhaseResult]]]] = []
        results: list[PhaseResult] = []

        for outcome in queue.run():
            if not isinstance(outcome.phase, PreparePlugin):
                continue

            if outcome.exc:
                outcome.logger.fail(str(outcome.exc))

                failed_tasks.append(outcome)
                continue

            if outcome.result:
                results += outcome.result

            self.preparations_applied += 1

        self._save_results(results)

        if failed_tasks:
            # TODO: needs a better message...
            raise tmt.utils.GeneralError(
                'prepare step failed',
                causes=[outcome.exc for outcome in failed_tasks if outcome.exc is not None],
            )

        self.info('')

        # Pull artifacts created in the plan data directory
        # if there was at least one plugin executed
        if self.phases() and guest_copies:
            sync_with_guests(
                self,
                'pull',
                PullTask(
                    logger=self._logger, guests=guest_copies, source=self.plan.data_directory
                ),
                self._logger,
            )

            # To separate "prepare" from "pull" queue visually
            self.info('')

        # Give a summary, update status and save
        self.summary()
        self.status('done')
        self.save()



"""
tmt's logging subsystem.

Adds a layer on top of Python's own :py:mod:`logging` subsystem. This layer implements the desired
verbosity and debug levels, colorization, formatting, verbosity inheritance and other features used
by tmt commands and code.

The main workhorses are :py:class:`Logger` instances. Each instance wraps a particular
:py:class:`logging.Logger` instance - usually there's a chain of such instances, with the root one
having console and logfile handlers attached. tmt's log verbosity/debug/quiet features are handled
on our side, with the use of :py:class:`logging.Filter` classes.

``Logger`` instances can be cloned and modified, to match various levels of tmt's runtime class
tree - ``tmt`` spawns a "root logger" from which a new one is cloned - and indented by one extra
level - for ``Run`` instance, and so on. This way, every object in tmt's hierarchy uses a given
logger, which may have its own specific settings, and, in the future, possibly also handlers for
special output channels.

While tmt recognizes several levels of verbosity (``-v``) and debugging (``-d``), all messages
emitted by :py:meth:`Logger.verbose` and :py:meth:`Logger.debug` use a single logging level,
``INFO`` or ``DEBUG``, respectively. The level of verbosity and debugging is then handled by a
special :py:class:`logging.Filter`` classes. This allows different levels when logging to console
but all-capturing log files while keeping implementation simple - the other option would be
managing handlers themselves, which would be very messy given the propagation of messages.
"""

import enum
import itertools
import logging
import os
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Protocol,
    TextIO,
    Union,
    cast,
)

from tmt._compat.pathlib import Path
from tmt._compat.warnings import deprecated
from tmt.container import container, simple_field

if TYPE_CHECKING:
    import tmt.cli
    import tmt.utils
    import tmt.utils.themes

# Log in workdir
LOG_FILENAME = 'log.txt'

# Hierarchy indent
INDENT = 4

DEFAULT_VERBOSITY_LEVEL = 0
DEFAULT_DEBUG_LEVEL = 0


class Topic(enum.Enum):
    KEY_NORMALIZATION = 'key-normalization'
    CLI_INVOCATIONS = 'cli-invocations'
    COMMAND_EVENTS = 'command-events'
    ADJUST_DECISIONS = 'adjust-decisions'
    HELP_RENDERING = 'help-rendering'


DEFAULT_TOPICS: set[Topic] = set()


LABEL_FORMAT = '[{label}]'


LoggableValue = Union[
    str,
    dict[str, Any],
    int,
    bool,
    float,
    'tmt.utils.Environment',
    'tmt.utils.FmfContext',
    'tmt.utils.Path',
    'tmt.utils.Command',
    'tmt.utils.ShellScript',
]


# TODO: this is an ugly hack, removing colors after they have been added...
# Wouldn't it be better to not add them at first place?
#
# This is needed to deal with the code that colorizes just part of the message, like
# tmt.result.Result outcomes: these are colorized, then merged with the number
# of such outcomes, for example, and the string is handed over to logging method.
# When colors are *not* to be applied, it's too late because colors have been
# applied already. Something to fix...
def _dont_decolorize(s: str) -> str:
    return s


def create_decolorizer(apply_colors: bool) -> Callable[[str], str]:
    if apply_colors:
        return _dont_decolorize

    import tmt.utils

    return tmt.utils.remove_color


def _debug_level_from_global_envvar() -> int:
    import tmt.utils

    raw_value = os.getenv('TMT_DEBUG', None)

    if raw_value is None:
        return 0

    try:
        return int(raw_value)

    except ValueError:
        raise tmt.utils.GeneralError(f"Invalid debug level '{raw_value}', use an integer.")


def decide_colorization(no_color: bool, force_color: bool) -> tuple[bool, bool]:
    """
    Decide whether the output and logging should be colorized.

    Based on values of CLI options, environment variables and output stream
    properties, a colorization setup is decided. The following inputs are
    evaluated, in this order:

    * if either of the ``--no-color`` CLI option, ``NO_COLOR`` or
        ``TMT_NO_COLOR`` environment variables are set, colorization would be
        disabled.
    * if either of the ``--force-color`` CLI option or ``TMT_FORCE_COLOR``
        environment variable are set, colorization would be forcefully
        enabled.

    If none of the situations above happened, colorization would be enabled for
    output and logging based on their respective stream TTY status. Output is
    sent to standard output, logging then to standard error output,
    colorization would then be the outcome of stream's :py:meth:`file.isatty`
    method.

    .. note::

       Be aware that "forced enable" is stronger than "forced disable". If
       ``--force-color`` or ``TMT_FORCE_COLOR`` are set, colors will be enabled
       despite any disabling options or environment variables.

    .. note::

       All inputs with the exception of ``isatty`` result control both types of
       output, regular output and logging, and applies to both of them. Only
       ``isatty`` outcome is specific for each type, and may result in one
       output type dropping colors while the other would be colorized.

    :param no_color: value of the ``--no-color`` CLI option.
    :param force_color: value of the `--force-color`` CLI option.
    :returns: a tuple of two booleans, one for output colorization, the other
        for logging colorization.
    """

    # Default values: assume colors & unicorns everywhere.
    apply_colors_output = apply_colors_logging = True

    # Enforce colors if `--force-color` was used, or `TMT_FORCE_COLOR` envvar is set.
    if force_color or 'TMT_FORCE_COLOR' in os.environ:
        apply_colors_output = apply_colors_logging = True

    # Disable coloring if `--no-color` was used, or `NO_COLOR` or `TMT_NO_COLOR` envvar is set.
    elif no_color or 'NO_COLOR' in os.environ or 'TMT_NO_COLOR' in os.environ:
        apply_colors_output = apply_colors_logging = False

    # Autodetection, disable colors when not talking to a terminal.
    else:
        apply_colors_output = sys.stdout.isatty()
        apply_colors_logging = sys.stderr.isatty()

    return apply_colors_output, apply_colors_logging


def render_labels(labels: list[str]) -> str:
    if not labels:
        return ''

    from tmt.utils.themes import style

    return ''.join(
        # TODO: color here is questionable - it will be removed, but I'd rather not
        # add it at first place, and it should be configurable.
        style(LABEL_FORMAT.format(label=label), fg='cyan')
        for label in labels
    )


def indent(
    key: str,
    value: Optional[LoggableValue] = None,
    color: 'tmt.utils.themes.Style' = None,
    level: int = 0,
    labels: Optional[list[str]] = None,
    labels_padding: int = 0,
) -> str:
    """
    Indent a key/value message.

    If both ``key`` and ``value`` are specified, ``{key}: {value}``
    message is rendered. Otherwise, just ``key`` is used alone. If
    ``value`` contains multiple lines, each but the very first line is
    indented by one extra level.

    :param value: optional value to print at right side of ``key``.
    :param color: optional color to apply on ``key``.
    :param level: number of indentation levels. Each level is indented
                  by :py:data:`INDENT` spaces.
    :param labels: optional list of strings to prepend to each message.
        Each item would be wrapped within square brackets (``[foo] message...``).
    :param labels_padding: if set, rendered labels would be padded to this
        length.
    """

    from tmt.utils.themes import style

    indent = ' ' * INDENT * level

    # Colorize
    key = style(key, style=color)

    # Prepare prefix if labels provided
    prefix = render_labels(labels).ljust(labels_padding) + ' ' if labels else ''

    # Handle key only
    if value is None:
        return f'{prefix}{indent}{key}'

    # Key + non-string values
    if not isinstance(value, str):
        from tmt.utils import format_value

        value = format_value(value, wrap=False)

    # If there's just a single line (or less...), emit just that line,
    # with prefix and indentation, of course.
    lines = value.splitlines()
    if len(lines) <= 1:
        return f'{prefix}{indent}{key}: {value}'

    # If we have multiple lines to emit, a key is emitted on its own line,
    # and all lines of the value are emitted below the key, all with an
    # extra bit of indentation ("deeper").
    deeper = ' ' * INDENT

    return f'{prefix}{indent}{key}:\n' + '\n'.join(
        f'{prefix}{indent}{deeper}{line}' for line in lines
    )


@container
class LogRecordDetails:
    """
    tmt's log message components attached to log records
    """

    key: str
    value: Optional[LoggableValue] = None

    color: 'tmt.utils.themes.Style' = None
    shift: int = 0

    logger_labels: list[str] = simple_field(default_factory=list[str])
    logger_labels_padding: int = 0

    logger_verbosity_level: int = 0
    message_verbosity_level: Optional[int] = None

    logger_debug_level: int = 0
    message_debug_level: Optional[int] = None

    logger_quiet: bool = False
    ignore_quietness: bool = False

    logger_topics: set[Topic] = simple_field(default_factory=set[Topic])
    message_topic: Optional[Topic] = None


class LogfileHandler(logging.FileHandler):
    #: Paths of all log files to which ``LogfileHandler`` was attached.
    emitting_to: list[Path] = []

    def __init__(self, filepath: 'tmt.utils.Path') -> None:
        super().__init__(filepath, mode='a')

        LogfileHandler.emitting_to.append(filepath)


# ignore[type-arg]: StreamHandler is a generic type, but such expression would be incompatible
# with older Python versions. Since it's not critical to mark the handler as "str only", we can
# ignore the issue for now.
class ConsoleHandler(logging.StreamHandler):  # type: ignore[type-arg]
    pass


class _Formatter(logging.Formatter):
    def __init__(self, fmt: str, apply_colors: bool = False) -> None:
        super().__init__(fmt, datefmt='%H:%M:%S')

        self.apply_colors = apply_colors

        self._decolorize = create_decolorizer(apply_colors)

    def format(self, record: logging.LogRecord) -> str:
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)

        # When message already exists, do nothing - it either some other logging subsystem,
        # or tmt's own, already rendered message.
        if hasattr(record, 'message'):
            pass

        # Otherwise render the message.
        elif record.msg and record.args:
            record.message = record.msg % record.args

        else:
            record.message = record.msg

        # Original code from Formatter.format() - hard to inherit when overriding
        # Formatter.format()...
        s = self._decolorize(self.formatMessage(record))
        # SIM102: Use a single `if` statement instead of nested `if` statements. Keeping for
        # readability.
        if record.exc_info:  # noqa: SIM102
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != "\n":
                s = s + "\n"
            s = s + self.formatStack(record.stack_info)
        return s


class LogfileFormatter(_Formatter):
    def __init__(self) -> None:
        super().__init__('%(asctime)s %(message)s', apply_colors=False)


class ConsoleFormatter(_Formatter):
    def __init__(self, apply_colors: bool = True, show_timestamps: bool = False) -> None:
        super().__init__(
            '%(asctime)s %(message)s' if show_timestamps else '%(message)s',
            apply_colors=apply_colors,
        )


class VerbosityLevelFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno != logging.INFO:
            return True

        details: Optional[LogRecordDetails] = getattr(record, 'details', None)

        if details is None:
            return True

        if details.message_verbosity_level is None:
            return True

        return details.logger_verbosity_level >= details.message_verbosity_level


class DebugLevelFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno != logging.DEBUG:
            return True

        details: Optional[LogRecordDetails] = getattr(record, 'details', None)

        if details is None:
            return True

        if details.message_debug_level is None:
            return True

        return details.logger_debug_level >= details.message_debug_level


class QuietnessFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno not in (logging.DEBUG, logging.INFO, logging.WARNING):
            return True

        details: Optional[LogRecordDetails] = getattr(record, 'details', None)

        if details is None:
            return False

        if not details.logger_quiet:
            return True

        if details.ignore_quietness:
            return True

        return False


class TopicFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno not in (logging.DEBUG, logging.INFO):
            return True

        details: Optional[LogRecordDetails] = getattr(record, 'details', None)

        if details is None:
            return False

        if details.message_topic is None:
            return True

        if details.message_topic in details.logger_topics:
            return True

        return False


class LoggingFunction(Protocol):
    def __call__(
        self,
        key: str,
        value: Optional[str] = None,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
        level: int = 1,
        topic: Optional[Topic] = None,
    ) -> None:
        pass


class Logger:
    """
    A logging entry point, representing a certain level of verbosity and handlers.

    Provides actual logging methods plus methods for managing verbosity levels
    and handlers.
    """

    def __init__(
        self,
        actual_logger: logging.Logger,
        base_shift: int = 0,
        labels: Optional[list[str]] = None,
        labels_padding: int = 0,
        verbosity_level: int = DEFAULT_VERBOSITY_LEVEL,
        debug_level: int = DEFAULT_DEBUG_LEVEL,
        quiet: bool = False,
        topics: Optional[set[Topic]] = None,
        apply_colors_output: bool = True,
        apply_colors_logging: bool = True,
    ) -> None:
        """
        Create a ``Logger`` instance with given verbosity levels.

        :param actual_logger: a :py:class:`logging.Logger` instance, the raw logger
            to use for logging.
        :param base_shift: shift applied to all messages processed by this logger.
        :param labels_padding: if set, rendered labels would be padded to this
            length.
        :param verbosity_level: desired verbosity level, usually derived from ``-v``
            command-line option.
        :param debug_level: desired debugging level, usually derived from ``-d``
            command-line option.
        :param quiet: if set, all messages would be suppressed, with the exception of
            warnings (:py:meth:`warn`), errors (:py:meth:`fail`) and messages emitted
            with :py:meth:`print`.
        """

        self._logger = actual_logger

        self._base_shift = base_shift

        self._child_id_counter = itertools.count()

        self.labels = labels or []
        self.labels_padding = labels_padding

        self.verbosity_level = verbosity_level
        self.debug_level = debug_level
        self.quiet = quiet
        self.topics = topics or DEFAULT_TOPICS

        self.apply_colors_output = apply_colors_output
        self.apply_colors_logging = apply_colors_logging

        self._decolorize_output = create_decolorizer(apply_colors_output)

    def __repr__(self) -> str:
        return (
            f'<Logger:'
            f' name={self._logger.name}'
            f' verbosity={self.verbosity_level}'
            f' debug={self.debug_level}'
            f' quiet={self.quiet}'
            f' topics={self.topics}'
            f' apply_colors_output={self.apply_colors_output}'
            f' apply_colors_logging={self.apply_colors_logging}'
            f'>'
        )

    @property
    def apply_colors_output(self) -> bool:
        return self._apply_colors_output

    @apply_colors_output.setter
    def apply_colors_output(self, value: bool) -> None:
        self._apply_colors_output = value

        self._decolorize_output = create_decolorizer(self._apply_colors_output)

    @property
    def labels_span(self) -> int:
        """
        Length of rendered labels
        """

        return len(render_labels(self.labels))

    @staticmethod
    def _normalize_logger(logger: logging.Logger) -> logging.Logger:
        """
        Reset properties of a given :py:class:`logging.Logger` instance
        """

        logger.propagate = True
        logger.level = logging.DEBUG

        logger.handlers = []

        return logger

    def clone(self) -> 'Logger':
        """
        Create a copy of this logger instance.

        All its settings are propagated to new instance. Settings are **not** shared,
        and may be freely modified after cloning without affecting the other logger.
        """

        return Logger(
            self._logger,
            base_shift=self._base_shift,
            labels=self.labels[:],
            labels_padding=self.labels_padding,
            verbosity_level=self.verbosity_level,
            debug_level=self.debug_level,
            quiet=self.quiet,
            topics=self.topics,
            apply_colors_output=self.apply_colors_output,
            apply_colors_logging=self.apply_colors_logging,
        )

    def descend(
        self,
        logger_name: Optional[str] = None,
        extra_shift: int = 1,
    ) -> 'Logger':
        """
        Create a copy of this logger instance, but with a new raw logger.

        New :py:class:`logging.Logger` instance is created from our raw logger, forming a
        parent/child relationship between them, and it's then wrapped with ``Logger`` instance.
        Settings of this logger are copied to new one, with the exception of ``base_shift``
        which is increased by one, effectively indenting all messages passing through new logger.

        :param logger_name: optional name for the underlying :py:class:`logging.Logger` instance.
            Useful for debugging. If not set, a generic one is created.
        :param extra_shift: by how many extra levels should messages be indented by new logger.
        """

        logger_name = logger_name or f'logger{next(self._child_id_counter)}'
        actual_logger = self._normalize_logger(self._logger.getChild(logger_name))

        return Logger(
            actual_logger,
            base_shift=self._base_shift + extra_shift,
            labels=self.labels[:],
            labels_padding=self.labels_padding,
            verbosity_level=self.verbosity_level,
            debug_level=self.debug_level,
            quiet=self.quiet,
            topics=self.topics,
            apply_colors_output=self.apply_colors_output,
            apply_colors_logging=self.apply_colors_logging,
        )

    def add_logfile_handler(self, filepath: 'tmt.utils.Path') -> None:
        """
        Attach a log file handler to this logger
        """

        handler = LogfileHandler(filepath)

        handler.setFormatter(LogfileFormatter())

        handler.addFilter(TopicFilter())

        self._logger.addHandler(handler)

    def add_console_handler(self, show_timestamps: bool = False) -> None:
        """
        Attach console handler to this logger.

        :param show_timestamps: when set, emitted messages would include
            the time.
        """

        handler = ConsoleHandler(stream=sys.stderr)

        handler.setFormatter(
            ConsoleFormatter(
                apply_colors=self.apply_colors_logging, show_timestamps=show_timestamps
            )
        )

        handler.addFilter(VerbosityLevelFilter())
        handler.addFilter(DebugLevelFilter())
        handler.addFilter(QuietnessFilter())
        handler.addFilter(TopicFilter())

        self._logger.addHandler(handler)

    def apply_verbosity_options(
        self,
        cli_invocation: Optional['tmt.cli.CliInvocation'] = None,
        **kwargs: Any,
    ) -> 'Logger':
        """
        Update logger's settings to match given CLI options.

        Use this method to update logger's settings after :py:meth:`Logger.descend` call,
        to reflect options given to a tmt subcommand.
        """

        actual_kwargs: dict[str, Any] = {}

        if cli_invocation is not None:
            actual_kwargs = cli_invocation.options

        actual_kwargs.update(kwargs)

        verbosity_level = cast(Optional[int], actual_kwargs.get('verbose', None))
        if verbosity_level is None or verbosity_level == 0:
            pass

        else:
            self.verbosity_level = verbosity_level

        debug_level_from_global_envvar = _debug_level_from_global_envvar()

        if debug_level_from_global_envvar not in (None, 0):
            self.debug_level = debug_level_from_global_envvar

        else:
            debug_level_from_option = cast(Optional[int], actual_kwargs.get('debug', None))

            if debug_level_from_option is None or debug_level_from_option == 0:
                pass

            else:
                self.debug_level = debug_level_from_option

        quietness_level = actual_kwargs.get('quiet', False)

        if quietness_level is True:
            self.quiet = quietness_level

        topic_specs = actual_kwargs.get('log_topic', [])

        for topic_spec in topic_specs:
            try:
                self.topics.add(Topic(topic_spec))

            except Exception:
                import tmt.utils

                raise tmt.utils.GeneralError(
                    f'Logging topic "{topic_spec}" is invalid.'
                    f" Possible choices are {', '.join(topic.value for topic in Topic)}"
                )

        return self

    @classmethod
    def create(
        cls,
        actual_logger: Optional[logging.Logger] = None,
        apply_colors_output: bool = True,
        apply_colors_logging: bool = True,
        **verbosity_options: Any,
    ) -> 'Logger':
        """
        Create a (root) tmt logger.

        This method has a very limited set of use cases:

        * CLI bootstrapping right after tmt started.
        * Unit tests of code that requires logger as one of its inputs.
        * 3rd party apps treating tmt as a library, i.e. when they wish tmt to
          use their logger instead of tmt's default one.

        :param actual_logger: a :py:class:`logging.Logger` instance to wrap.
            If not set, a default logger named ``tmt`` is created.
        """

        actual_logger = actual_logger or cls._normalize_logger(logging.getLogger('tmt'))

        return Logger(
            actual_logger,
            apply_colors_output=apply_colors_output,
            apply_colors_logging=apply_colors_logging,
        ).apply_verbosity_options(**verbosity_options)

    def _log(
        self,
        level: int,
        details: LogRecordDetails,
        message: str = '',
    ) -> None:
        """
        Emit a log record describing the message and related properties.

        This method converts tmt's specific logging approach, with keys, values, colors
        and shifts, to :py:class:`logging.LogRecord` instances compatible with :py:mod:`logging`
        workflow and carrying extra information for our custom filters and handlers.
        """

        details.logger_labels = self.labels
        details.logger_labels_padding = self.labels_padding

        details.logger_verbosity_level = self.verbosity_level
        details.logger_debug_level = self.debug_level
        details.logger_quiet = self.quiet
        details.logger_topics = self.topics

        details.shift = details.shift + self._base_shift

        if not message:
            message = indent(
                details.key,
                value=details.value,
                # Always apply colors - message can be decolorized later.
                color=details.color,
                level=details.shift,
                labels=self.labels,
                labels_padding=self.labels_padding,
            )

        self._logger._log(level, message, (), extra={'details': details})

    def print_format(
        self,
        text: str,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
    ) -> str:
        """
        Format the given text in a way suitable for :py:meth:`print`
        """

        text = indent(
            text,
            # Always apply colors - message can be decolorized later.
            color=color,
            level=shift + self._base_shift,
            labels=self.labels,
            labels_padding=self.labels_padding,
        )

        return self._decolorize_output(text)

    def print(
        self,
        text: str,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
        file: Optional[TextIO] = None,
    ) -> None:
        file = file or sys.stdout

        print(self.print_format(text, color=color, shift=shift), file=file)

    def info(
        self,
        key: str,
        value: Optional[LoggableValue] = None,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
    ) -> None:
        self._log(logging.INFO, LogRecordDetails(key=key, value=value, color=color, shift=shift))

    def verbose(
        self,
        key: str,
        value: Optional[LoggableValue] = None,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
        level: int = 1,
        topic: Optional[Topic] = None,
    ) -> None:
        self._log(
            logging.INFO,
            LogRecordDetails(
                key=key,
                value=value,
                color=color,
                shift=shift,
                message_verbosity_level=level,
                message_topic=topic,
            ),
        )

    def debug(
        self,
        key: str,
        value: Optional[LoggableValue] = None,
        color: 'tmt.utils.themes.Style' = None,
        shift: int = 0,
        level: int = 1,
        topic: Optional[Topic] = None,
    ) -> None:
        self._log(
            logging.DEBUG,
            LogRecordDetails(
                key=key,
                value=value,
                color=color,
                shift=shift,
                message_debug_level=level,
                message_topic=topic,
            ),
        )

    def warning(
        self,
        message: str,
        shift: int = 0,
    ) -> None:
        self._log(
            logging.WARNING,
            LogRecordDetails(key='warn', value=message, color='yellow', shift=shift),
        )

    @deprecated("Use Logger.warning instead")
    def warn(
        self,
        message: str,
        shift: int,
    ) -> None:
        return self.warning(message, shift)

    def fail(
        self,
        message: str,
        shift: int = 0,
    ) -> None:
        self._log(
            logging.ERROR, LogRecordDetails(key='fail', value=message, color='red', shift=shift)
        )

    _bootstrap_logger: Optional['Logger'] = None

    @classmethod
    def get_bootstrap_logger(cls) -> 'Logger':
        """
        Create a logger designed for tmt startup time.

        .. warning::

            This logger has a **very** limited use case span, i.e.
            before tmt can digest its command-line options and create a
            proper logger. This happens inside
            :py:func:`tmt.cli._root.main` function, but there are some
            actions taken by tmt code before this function is called by
            Click, actions that need to emit logging messages. Using it
            anywhere outside of this brief time in tmt's runtime should
            be ruled out.
        """

        if cls._bootstrap_logger is None:
            # Stay away of our future main logger
            actual_logger = Logger._normalize_logger(logging.getLogger('_tmt_bootstrap'))

            cls._bootstrap_logger = Logger.create(actual_logger=actual_logger)
            cls._bootstrap_logger.add_console_handler()

        return cls._bootstrap_logger



import collections
import functools
import itertools
import os
import platform
import re
import threading
import types
from collections.abc import Iterator
from string import Template
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import click
import pint
import requests

import tmt
import tmt.hardware
import tmt.log
import tmt.steps
import tmt.steps.provision
import tmt.utils
import tmt.utils.wait
from tmt.container import container, field
from tmt.utils import (
    WORKDIR_ROOT,
    Command,
    Path,
    ProvisionError,
    ShellScript,
    configure_constant,
    retry_session,
)
from tmt.utils.wait import Deadline, Waiting

if TYPE_CHECKING:
    import tmt.base
    from tmt.hardware import Size


libvirt: Optional[types.ModuleType] = None
testcloud: Optional[types.ModuleType] = None

# To silence mypy
DomainConfiguration: Any
Workarounds: Any
X86_64ArchitectureConfiguration: Any
AArch64ArchitectureConfiguration: Any
S390xArchitectureConfiguration: Any
Ppc64leArchitectureConfiguration: Any
SystemNetworkConfiguration: Any
UserNetworkConfiguration: Any
QCow2StorageDevice: Any
RawStorageDevice: Any
TPMConfiguration: Any


def import_testcloud(logger: tmt.log.Logger) -> None:
    """
    Import testcloud module only when needed
    """

    global testcloud
    global libvirt
    global Workarounds
    global DomainConfiguration
    global X86_64ArchitectureConfiguration
    global AArch64ArchitectureConfiguration
    global S390xArchitectureConfiguration
    global Ppc64leArchitectureConfiguration
    global SystemNetworkConfiguration
    global UserNetworkConfiguration
    global QCow2StorageDevice
    global RawStorageDevice
    global TPMConfiguration
    try:
        import libvirt
        import testcloud.image
        import testcloud.instance
        import testcloud.util
        from testcloud.domain_configuration import (
            AArch64ArchitectureConfiguration,
            DomainConfiguration,
            Ppc64leArchitectureConfiguration,
            QCow2StorageDevice,
            RawStorageDevice,
            S390xArchitectureConfiguration,
            SystemNetworkConfiguration,
            TPMConfiguration,
            UserNetworkConfiguration,
            X86_64ArchitectureConfiguration,
        )
        from testcloud.workarounds import Workarounds
    except ImportError as error:
        from tmt.utils.hints import print_hints

        print_hints('provision/virtual.testcloud', logger=logger)

        raise ProvisionError('Could not import testcloud package.') from error

    # Version-aware TPM configuration is added in
    # https://pagure.io/testcloud/c/89f1c024ca829543de7f74f89329158c6dee3d83
    global TPM_CONFIG_ALLOWS_VERSIONS
    TPM_CONFIG_ALLOWS_VERSIONS = hasattr(TPMConfiguration(), 'version')


# Testcloud cache to our tmt's workdir root
TESTCLOUD_DATA = (
    Path(os.environ['TMT_WORKDIR_ROOT']) if os.getenv('TMT_WORKDIR_ROOT') else WORKDIR_ROOT
) / 'testcloud'
TESTCLOUD_IMAGES = TESTCLOUD_DATA / 'images'

TESTCLOUD_WORKAROUNDS: list[str] = []  # A list of commands to be executed during guest boot up

# Userdata for cloud-init
# TODO: Explore migration to jinja from string.Template
USER_DATA = """#cloud-config
chpasswd:
  list: |
    ${user_name}:${password}
  expire: false
users:
  - default
  - name: ${user_name}
ssh_authorized_keys:
  - ${public_key}
ssh_pwauth: true
disable_root: false
runcmd:
${runcommands}
"""

COREOS_DATA = """variant: fcos
version: 1.4.0
passwd:
  users:
    - name: ${user_name}
      ssh_authorized_keys:
        - ${public_key}
storage:
  files:
    - path: /etc/ssh/sshd_config.d/20-enable-root-login.conf
      mode: 0644
      contents:
        inline: |
          # CoreOS disables root SSH login by default.
          # Enable it.
          # This file must sort before 40-rhcos-defaults.conf.
          PermitRootLogin yes
"""

# VM defaults
#: How many seconds to wait for a VM to start.
#: This is the default value tmt would use unless told otherwise.
DEFAULT_BOOT_TIMEOUT: int = 2 * 60

#: How many seconds to wait for a VM to start.
#: This is the effective value, combining the default and optional envvar,
#: ``TMT_BOOT_TIMEOUT``.
BOOT_TIMEOUT: int = configure_constant(DEFAULT_BOOT_TIMEOUT, 'TMT_BOOT_TIMEOUT')

#: How many seconds to wait for a connection to succeed after guest boot.
#: This is the default value tmt would use unless told otherwise.
DEFAULT_CONNECT_TIMEOUT = 2 * 60

#: How many seconds to wait for a connection to succeed after guest boot.
#: This is the effective value, combining the default and optional envvar,
#: ``TMT_CONNECT_TIMEOUT``.
CONNECT_TIMEOUT: int = configure_constant(DEFAULT_CONNECT_TIMEOUT, 'TMT_CONNECT_TIMEOUT')

#: How many times should the timeouts be multiplied in kvm-less cases.
#: These include emulating a different architecture than the host,
#: some nested virtualization cases, and hosts with degraded virt caps.
NON_KVM_TIMEOUT_COEF = 10  # times

# SSH key type, set None for ssh-keygen default one
SSH_KEYGEN_TYPE = "ecdsa"

DEFAULT_USER = 'root'
DEFAULT_CPU_COUNT = 2
DEFAULT_MEMORY: 'Size' = tmt.hardware.UNITS('2048 MB')
DEFAULT_DISK: 'Size' = tmt.hardware.UNITS('40 GB')
DEFAULT_IMAGE = 'fedora'
DEFAULT_CONNECTION = 'session'
DEFAULT_ARCH = platform.machine()

# Version-aware TPM configuration is added in
# https://pagure.io/testcloud/c/89f1c024ca829543de7f74f89329158c6dee3d83
#: If set, ``testcloud`` TPM configuration accepts TPM version as a parameter.
TPM_CONFIG_ALLOWS_VERSIONS: bool = False

#: List of operators supported for ``tpm.version`` HW requirement.
TPM_VERSION_ALLOWED_OPERATORS: tuple[tmt.hardware.Operator, ...] = (
    tmt.hardware.Operator.EQ,
    tmt.hardware.Operator.GTE,
    tmt.hardware.Operator.LTE,
)

#: TPM versions supported by the plugin. The key is :py:const:`TPM_CONFIG_ALLOWS_VERSIONS`.
TPM_VERSION_SUPPORTED_VERSIONS = {
    True: ['2.0', '2', '1.2'],
    # This is the default version used by testcloud before version became
    # an input parameter of TPM configuration.
    False: ['2.0', '2'],
}


def normalize_memory_size(
    key_address: str,
    value: Any,
    logger: tmt.log.Logger,
) -> Optional['Size']:
    """
    Normalize memory size.

    As of now, it's just a simple integer with implicit unit, ``MB``.
    """

    if value is None:
        return None

    if isinstance(value, pint.Quantity):
        return value

    if isinstance(value, str):
        try:
            magnitude = int(value)

        except ValueError:
            return tmt.hardware.UNITS(value)

        return tmt.hardware.UNITS(f'{magnitude} MB')

    if isinstance(value, int):
        return tmt.hardware.UNITS(f'{value} MB')

    raise tmt.utils.NormalizationError(key_address, value, 'an integer')


def normalize_disk_size(key_address: str, value: Any, logger: tmt.log.Logger) -> Optional['Size']:
    """
    Normalize disk size.

    As of now, it's just a simple integer with implicit unit, ``GB``.
    """

    if value is None:
        return None

    if isinstance(value, pint.Quantity):
        return value

    if isinstance(value, str):
        try:
            magnitude = int(value)

        except ValueError:
            return tmt.hardware.UNITS(value)

        return tmt.hardware.UNITS(f'{magnitude} GB')

    if isinstance(value, int):
        return tmt.hardware.UNITS(f'{value} GB')

    raise tmt.utils.NormalizationError(key_address, value, 'an integer')


def _report_hw_requirement_support(constraint: tmt.hardware.Constraint) -> bool:
    components = constraint.expand_name()

    if components.name == 'memory' and constraint.operator in (
        tmt.hardware.Operator.EQ,
        tmt.hardware.Operator.GTE,
        tmt.hardware.Operator.LTE,
    ):
        return True

    if (
        components.name == 'cpu'
        and components.child_name == 'processors'
        and constraint.operator
        in (
            tmt.hardware.Operator.EQ,
            tmt.hardware.Operator.LTE,
            tmt.hardware.Operator.GTE,
            tmt.hardware.Operator.NEQ,
            tmt.hardware.Operator.LT,
            tmt.hardware.Operator.GT,
        )
    ):
        return True

    if (
        components.name == 'disk'
        and components.child_name == 'size'
        and constraint.operator
        in (tmt.hardware.Operator.EQ, tmt.hardware.Operator.GTE, tmt.hardware.Operator.LTE)
    ):
        return True

    if (
        components.name == 'tpm'
        and components.child_name == 'version'
        and constraint.value in TPM_VERSION_SUPPORTED_VERSIONS[TPM_CONFIG_ALLOWS_VERSIONS]
        and constraint.operator in TPM_VERSION_ALLOWED_OPERATORS
    ):
        return True

    return False


@container
class TestcloudGuestData(tmt.steps.provision.GuestSshData):
    # Override parent class with our defaults
    user: str = field(
        default=DEFAULT_USER,
        option=('-u', '--user'),
        metavar='USERNAME',
        help='Username to use for all guest operations.',
    )

    image: str = field(
        default=DEFAULT_IMAGE,
        option=('-i', '--image'),
        metavar='IMAGE',
        help="""
             Select image to be used. Provide a short name, full path to a local file
             or a complete url.
             """,
    )
    memory: Optional['Size'] = field(
        default=cast(Optional['Size'], None),
        option=('-m', '--memory'),
        metavar='SIZE',
        help='Set available memory in MB, 2048 MB by default.',
        normalize=normalize_memory_size,
        serialize=lambda value: str(value) if value is not None else None,
        unserialize=lambda serialized: tmt.hardware.UNITS(serialized)
        if serialized is not None
        else None,
    )
    disk: Optional['Size'] = field(
        default=cast(Optional['Size'], None),
        option=('-D', '--disk'),
        metavar='SIZE',
        help='Specify disk size in GB, 10 GB by default.',
        normalize=normalize_disk_size,
        serialize=lambda value: str(value) if value is not None else None,
        unserialize=lambda serialized: tmt.hardware.UNITS(serialized)
        if serialized is not None
        else None,
    )
    connection: str = field(
        default=DEFAULT_CONNECTION,
        option=('-c', '--connection'),
        choices=['session', 'system'],
        help="What session type to use, 'session' by default.",
    )
    arch: str = field(
        default=DEFAULT_ARCH,
        option=('-a', '--arch'),
        choices=['x86_64', 'aarch64', 's390x', 'ppc64le'],
        help="What architecture to virtualize, host arch by default.",
    )

    list_local_images: bool = field(
        default=False,
        option='--list-local-images',
        is_flag=True,
        help="List locally available images.",
    )

    image_url: Optional[str] = field(
        default=None,
        internal=True,
    )
    instance_name: Optional[str] = field(
        default=None,
        internal=True,
    )

    # TODO: custom handling for two fields - when the formatting moves into
    # field(), this should not be necessary.
    def show(
        self,
        *,
        keys: Optional[list[str]] = None,
        verbose: int = 0,
        logger: tmt.log.Logger,
    ) -> None:
        keys = keys or list(self.keys())
        super_keys = [key for key in keys if key not in ('memory', 'disk')]

        super().show(keys=super_keys, verbose=verbose, logger=logger)

        # TODO: find formatting that would show "MB" instead of "megabyte"
        # https://github.com/teemtee/tmt/issues/2410
        logger.info('memory', f'{(self.memory or DEFAULT_MEMORY).to("MB")}', 'green')
        logger.info('disk', f'{(self.disk or DEFAULT_DISK).to("GB")}', 'green')


@container
class ProvisionTestcloudData(TestcloudGuestData, tmt.steps.provision.ProvisionStepData):
    pass


def _apply_hw_tpm(
    hardware: Optional[tmt.hardware.Hardware],
    domain: 'DomainConfiguration',
    logger: tmt.log.Logger,
) -> None:
    """
    Apply ``tpm`` constraint to given VM domain
    """

    domain.tpm_configuration = None

    if not hardware or not hardware.constraint:
        logger.debug('tpm.version', "not included because of no constraints", level=4)

        return

    variant = hardware.constraint.variant()

    tpm_constraints = [
        constraint
        for constraint in variant
        if isinstance(constraint, tmt.hardware.TextConstraint)
        and constraint.expand_name().name == 'tpm'
        and constraint.expand_name().child_name == 'version'
    ]

    if not tpm_constraints:
        logger.debug(
            'tpm.version', "not included because of no 'tpm.version' constraints", level=4
        )

        return

    for constraint in tpm_constraints:
        if constraint.operator not in TPM_VERSION_ALLOWED_OPERATORS:
            logger.warning(
                f"Cannot apply hardware requirement '{constraint}', operator not supported."
            )
            return

        if constraint.value not in TPM_VERSION_SUPPORTED_VERSIONS[TPM_CONFIG_ALLOWS_VERSIONS]:
            logger.warning(
                f"Cannot apply hardware requirement '{constraint}', TPM version not supported."
            )
            return

        logger.debug(
            'tpm.version', f"set to '{constraint.value}' because of '{constraint}'", level=4
        )

        if TPM_CONFIG_ALLOWS_VERSIONS:
            domain.tpm_configuration = TPMConfiguration(version=constraint.value)

        else:
            domain.tpm_configuration = TPMConfiguration()


def _apply_hw_disk_size(
    hardware: Optional[tmt.hardware.Hardware],
    domain: 'DomainConfiguration',
    logger: tmt.log.Logger,
) -> None:
    """
    Apply ``disk`` constraint to given VM domain
    """

    final_size: Size = DEFAULT_DISK

    def _generate_disk_filepaths() -> Iterator[Path]:
        """
        Generate paths to use for files representing VM storage
        """

        # Start with the path already decided by testcloud...
        yield Path(domain.local_disk)

        # ... and use it as a basis for remaining paths.
        for i in itertools.count(1, 1):
            yield Path(f'{domain.local_disk}.{i}')

    disk_filepath_generator = _generate_disk_filepaths()

    if not hardware or not hardware.constraint:
        logger.debug('disk[0].size', f"set to '{final_size}' because of no constraints", level=4)

        domain.storage_devices = [
            QCow2StorageDevice(
                str(next(disk_filepath_generator)), int(final_size.to('GB').magnitude)
            )
        ]

        return

    variant = hardware.constraint.variant()

    # Collect all `disk.size` constraints, ignore the rest.
    disk_size_constraints = [
        constraint
        for constraint in variant
        if isinstance(constraint, tmt.hardware.SizeConstraint)
        and constraint.expand_name().name == 'disk'
        and constraint.expand_name().child_name == 'size'
    ]

    if not disk_size_constraints:
        logger.debug(
            'disk[0].size', f"set to '{final_size}' because of no 'disk.size' constraints", level=4
        )

        domain.storage_devices = [
            QCow2StorageDevice(
                str(next(disk_filepath_generator)), int(final_size.to('GB').magnitude)
            )
        ]

        return

    # Now sort them into groups by their `peer_index`, i.e. `disk[0]`,
    # `disk[1]` and so on.
    by_peer_index: dict[int, list[tmt.hardware.SizeConstraint]] = collections.defaultdict(list)

    for constraint in disk_size_constraints:
        if constraint.operator not in (
            tmt.hardware.Operator.EQ,
            tmt.hardware.Operator.GTE,
            tmt.hardware.Operator.LTE,
        ):
            raise ProvisionError(
                f"Cannot apply hardware requirement '{constraint}', operator not supported."
            )

        components = constraint.expand_name()

        assert components.peer_index is not None  # narrow type

        by_peer_index[components.peer_index].append(constraint)

    # Process each disk and its constraints, construct the
    # corresponding storage device, and the last constraint wins
    # & sets its size.
    for peer_index in sorted(by_peer_index.keys()):
        final_size = DEFAULT_DISK

        for constraint in by_peer_index[peer_index]:
            logger.debug(
                f'disk[{peer_index}].size',
                f"set to '{constraint.value}' because of '{constraint}'",
                level=4,
            )

            final_size = constraint.value

        domain.storage_devices.append(
            QCow2StorageDevice(
                str(next(disk_filepath_generator)), int(final_size.to('GB').magnitude)
            )
        )


def _apply_hw_cpu_processors(
    hardware: Optional[tmt.hardware.Hardware],
    domain: 'DomainConfiguration',
    logger: tmt.log.Logger,
) -> None:
    """
    Apply ``cpu.processors`` constraint to given VM domain
    """

    domain.cpu_count = DEFAULT_CPU_COUNT

    if not hardware or not hardware.constraint:
        logger.debug('cpu.processors', "not included because of no constraints", level=4)

        return

    variant = hardware.constraint.variant()

    cpu_processors_constraints = [
        constraint
        for constraint in variant
        if isinstance(constraint, tmt.hardware.IntegerConstraint)
        and constraint.expand_name().name == 'cpu'
        and constraint.expand_name().child_name == 'processors'
    ]

    if not cpu_processors_constraints:
        logger.debug(
            'cpu.processors', "not included because of no 'cpu.processors' constraints", level=4
        )

        return

    for constraint in cpu_processors_constraints:
        if constraint.operator in (
            tmt.hardware.Operator.EQ,
            tmt.hardware.Operator.LTE,
            tmt.hardware.Operator.GTE,
        ):
            logger.debug(
                'cpu.processors', f"set to '{constraint.value}' because of '{constraint}'", level=4
            )

            domain.cpu_count = constraint.value

        elif (
            constraint.operator is tmt.hardware.Operator.NEQ
            and domain.cpu_count != constraint.value
        ):
            logger.debug(
                'cpu.processors',
                f"kept at '{constraint.value}' because of '{constraint}'",
                level=4,
            )

        elif constraint.operator is tmt.hardware.Operator.LT:
            logger.debug(
                'cpu.processors',
                f"set to '{constraint.value - 1}' because of '{constraint}'",
                level=4,
            )

            domain.cpu_count = constraint.value - 1

        elif constraint.operator is tmt.hardware.Operator.GT:
            logger.debug(
                'cpu.processors',
                f"set to '{constraint.value + 1}' because of '{constraint}'",
                level=4,
            )

            domain.cpu_count = constraint.value + 1

        else:
            raise ProvisionError(
                f"Cannot apply hardware requirement '{constraint}', operator not supported."
            )


class GuestTestcloud(tmt.GuestSsh):
    """
    Testcloud Instance

    The following keys are expected in the 'data' dictionary::

        image ...... qcov image name or url
        user ....... user name to log in
        memory ..... memory size for vm
        disk ....... disk size for vm
        connection . either session (default) or system, to be passed to qemu
        arch ....... architecture for the VM, host arch is the default
    """

    _data_class = TestcloudGuestData

    image: str
    image_url: Optional[str]
    instance_name: Optional[str]
    memory: Optional['Size']
    disk: Optional['Size']
    connection: str
    arch: str

    # Not to be saved, recreated from image_url/instance_name/... every
    # time guest is instantiated.
    # FIXME: ignore[name-defined]: https://github.com/teemtee/tmt/issues/1616
    _image: Optional['testcloud.image.Image'] = None  # type: ignore[name-defined]
    _instance: Optional['testcloud.instance.Instance'] = None  # type: ignore[name-defined]
    _domain: Optional[  # type: ignore[name-defined]
        'testcloud.domain_configuration.DomainConfiguration'
    ] = None

    #: The lock protects calls into the testcloud library. We suspect it might
    #: be unprepared for multi-threaded use. After the dust settles, we may
    #: remove the lock.
    _testcloud_lock = threading.Lock()

    @property
    def is_ready(self) -> bool:
        if self._instance is None:
            return False

        assert testcloud is not None
        assert libvirt is not None
        try:
            state = testcloud.instance._find_domain(self._instance.name, self._instance.connection)
            # Note the type of variable 'state' is 'Any'. Hence, we don't use:
            #     return state == 'running'
            # to avoid error from type checking.
            return bool(state == "running")
        except libvirt.libvirtError:
            return False

    @functools.cached_property
    def is_kvm(self) -> bool:
        # Is the combination of host-requested architecture kvm capable?
        return bool(self.arch == platform.machine() and Path('/dev/kvm').exists())

    @functools.cached_property
    def is_legacy_os(self) -> bool:
        assert testcloud is not None  # narrow post-import type
        assert self._image is not None  # narrow type

        # Is this el <= 7?
        return cast(bool, testcloud.util.needs_legacy_net(self._image.name))

    @functools.cached_property
    def is_coreos(self) -> bool:
        # Is this a CoreOS?
        return bool(re.search('coreos|rhcos', self.image.lower()))

    def _get_url(self, url: str, message: str) -> requests.Response:
        """
        Get url, retry when fails, return response
        """

        def try_get_url() -> requests.Response:
            try:
                with retry_session() as session:
                    response = session.get(url)

                if response.ok:
                    return response

            except requests.RequestException:
                pass
            finally:
                raise tmt.utils.wait.WaitingIncompleteError

        try:
            return Waiting(Deadline.from_seconds(CONNECT_TIMEOUT), tick=1).wait(
                try_get_url, self._logger
            )

        except tmt.utils.wait.WaitingTimedOutError:
            raise ProvisionError(f'Failed to {message} in {CONNECT_TIMEOUT}s.')

    def _guess_image_url(self, name: str) -> str:
        """
        Guess image url for given name
        """

        # Try to check if given url is a local file
        name_as_path = Path(name)
        if name_as_path.is_absolute() and name_as_path.is_file():
            return f'file://{name}'

        url: Optional[str] = None
        assert testcloud is not None

        with GuestTestcloud._testcloud_lock:
            try:
                url = testcloud.util.get_image_url(name.lower().strip(), self.arch)
            except Exception as error:
                raise ProvisionError("Could not get image url.") from error

        if not url:
            raise ProvisionError(f"Could not map '{name}' to compose.")
        return url

    def wake(self) -> None:
        """
        Wake up the guest
        """

        self.debug(f"Waking up testcloud instance '{self.instance_name}'.", level=2, shift=0)
        self.prepare_config()
        assert testcloud is not None
        self._image = testcloud.image.Image(self.image_url)
        if self.instance_name is None:
            raise ProvisionError(f"The instance name '{self.instance_name}' is invalid.")

        self._domain = DomainConfiguration(self.instance_name)
        self._apply_hw_arch(self._domain, self.is_kvm, self.is_legacy_os)

        # Is this a CoreOS?
        self._domain.coreos = self.is_coreos

        self._instance = testcloud.instance.Instance(
            domain_configuration=self._domain,
            image=self._image,
            desired_arch=self.arch,
            connection=f"qemu:///{self.connection}",
        )

    def prepare_ssh_key(self, key_type: Optional[str] = None) -> str:
        """
        Prepare ssh key for authentication
        """

        assert self.workdir is not None

        # Use existing key
        if self.key:
            self.debug("Extract public key from the provided private one.")
            command = Command("ssh-keygen", "-f", self.key[0], "-y")
            public_key = self._run_guest_command(command, silent=True).stdout
            assert public_key is not None
        # Generate new ssh key pair
        else:
            self.debug('Generating an ssh key.')
            key_name = f"id_{key_type if key_type is not None else 'rsa'}"
            self.key = [self.workdir / key_name]
            command = Command("ssh-keygen", "-f", self.key[0], "-N", "")
            if key_type is not None:
                command += Command("-t", key_type)
            self._run_guest_command(command, silent=True)
            self.verbose('key', self.key[0], 'green')
            public_key = (self.workdir / f'{key_name}.pub').read_text()

        return public_key

    def prepare_config(self) -> None:
        """
        Prepare common configuration
        """

        import_testcloud(self._logger)

        # Get configuration
        assert testcloud is not None
        self.config = testcloud.config.get_config()

        self.debug(f"testcloud version: {testcloud.__version__}")

        # Make sure download progress is disabled unless in debug mode,
        # so it does not spoil our logging
        self.config.DOWNLOAD_PROGRESS = self.debug_level > 2
        self.config.DOWNLOAD_PROGRESS_VERBOSE = False

        # Configure to tmt's storage directories
        self.config.DATA_DIR = TESTCLOUD_DATA
        self.config.STORE_DIR = TESTCLOUD_IMAGES

    def _combine_hw_memory(self) -> None:
        """
        Combine ``hardware`` with ``--memory`` option
        """

        if not self.hardware:
            self.hardware = tmt.hardware.Hardware.from_spec({})

        if self.memory is None:
            return

        memory_constraint = tmt.hardware.SizeConstraint.from_specification(
            'memory', str(self.memory)
        )

        self.hardware.and_(memory_constraint)

    def _combine_hw_disk_size(self) -> None:
        """
        Combine ``hardware`` with ``--disk`` option
        """

        if not self.hardware:
            self.hardware = tmt.hardware.Hardware.from_spec({})

        if self.disk is None:
            return

        disk_size_constraint = tmt.hardware.SizeConstraint.from_specification(
            'disk[0].size', str(self.disk)
        )

        self.hardware.and_(disk_size_constraint)

    def _apply_hw_memory(self, domain: 'DomainConfiguration') -> None:
        """
        Apply ``memory`` constraint to given VM domain
        """

        if not self.hardware or not self.hardware.constraint:
            self.debug('memory', f"set to '{DEFAULT_MEMORY}' because of no constraints", level=4)

            domain.memory_size = int(DEFAULT_MEMORY.to('kB').magnitude)

            return

        variant = self.hardware.constraint.variant()

        memory_constraints = [
            constraint
            for constraint in variant
            if isinstance(constraint, tmt.hardware.SizeConstraint)
            and constraint.expand_name().name == 'memory'
        ]

        if not memory_constraints:
            self.debug(
                'memory', f"set to '{DEFAULT_MEMORY}' because of no 'memory' constraints", level=4
            )

            domain.memory_size = int(DEFAULT_MEMORY.to('kB').magnitude)

            return

        for constraint in memory_constraints:
            if constraint.operator not in (
                tmt.hardware.Operator.EQ,
                tmt.hardware.Operator.GTE,
                tmt.hardware.Operator.LTE,
            ):
                raise ProvisionError(
                    f"Cannot apply hardware requirement '{constraint}', operator not supported."
                )

            self.debug('memory', f"set to '{constraint.value}' because of '{constraint}'", level=4)

            domain.memory_size = int(constraint.value.to('kB').magnitude)

    def _apply_hw_arch(self, domain: 'DomainConfiguration', kvm: bool, legacy_os: bool) -> None:
        if self.arch == "x86_64":
            domain.system_architecture = X86_64ArchitectureConfiguration(
                kvm=kvm,
                uefi=False,  # Configurable
                model="q35" if not legacy_os else "pc",
            )
        elif self.arch == "aarch64":
            domain.system_architecture = AArch64ArchitectureConfiguration(
                kvm=kvm,
                uefi=True,  # Always enabled
                model="virt",
            )
        elif self.arch == "ppc64le":
            domain.system_architecture = Ppc64leArchitectureConfiguration(
                kvm=kvm,
                uefi=False,  # Always disabled
                model="pseries",
            )
        elif self.arch == "s390x":
            domain.system_architecture = S390xArchitectureConfiguration(
                kvm=kvm,
                uefi=False,  # Always disabled
                model="s390-ccw-virtio",
            )
        else:
            raise tmt.utils.ProvisionError("Unknown architecture requested.")

    def start(self) -> None:
        """
        Start provisioned guest
        """

        if self.is_dry_run:
            return
        # Make sure required directories exist
        os.makedirs(TESTCLOUD_DATA, exist_ok=True)
        os.makedirs(TESTCLOUD_IMAGES, exist_ok=True)

        # Prepare config
        self.prepare_config()

        # Kick off image URL with the given image
        self.image_url = self.image

        # If image does not start with http/https/file, consider it a
        # mapping value and try to guess the URL
        if not re.match(r'^(?:https?|file)://.*', self.image_url):
            self.image_url = self._guess_image_url(self.image_url)
            self.debug(f"Guessed image url: '{self.image_url}'", level=3)

        # Initialize and prepare testcloud image
        assert testcloud is not None
        self._image = testcloud.image.Image(self.image_url)
        self.verbose('qcow', self._image.name, 'green')
        if not Path(self._image.local_path).exists():
            self.info('progress', 'downloading...', 'cyan')
        try:
            self._image.prepare()
        except FileNotFoundError as error:
            raise ProvisionError(f"Image '{self._image.local_path}' not found.") from error
        except (testcloud.exceptions.TestcloudPermissionsError, PermissionError) as error:
            raise ProvisionError(
                f"Failed to prepare the image. Check the '{TESTCLOUD_IMAGES}' "
                f"directory permissions."
            ) from error
        except KeyError as error:
            raise ProvisionError(f"Failed to prepare image '{self.image_url}'.") from error

        # Prepare hostname (get rid of possible unwanted characters)
        hostname = re.sub(r"[^a-zA-Z0-9\-]+", "-", self.name.lower()).strip("-")

        # Create instance
        self.instance_name = self._tmt_name()

        # Prepare DomainConfiguration object before Instance object
        self._domain = DomainConfiguration(self.instance_name)

        # Prepare Workarounds object
        self._workarounds = Workarounds(defaults=True)
        for cmd in TESTCLOUD_WORKAROUNDS:
            self._workarounds.add(cmd)

        # Process hardware and find a suitable HW properties
        self._domain.cpu_count = DEFAULT_CPU_COUNT

        self._combine_hw_memory()
        self._combine_hw_disk_size()

        if self.hardware:
            self.verbose('effective hardware', self.hardware.to_spec(), color='green')

            for line in self.hardware.format_variants():
                self._logger.debug('effective hardware', line, level=4)

        self._apply_hw_memory(self._domain)
        _apply_hw_cpu_processors(self.hardware, self._domain, self._logger)
        _apply_hw_disk_size(self.hardware, self._domain, self._logger)
        _apply_hw_tpm(self.hardware, self._domain, self._logger)

        self.debug('final domain memory', str(self._domain.memory_size))
        self.debug('final domain root disk size', str(self._domain.storage_devices[0].size))

        for i, device in enumerate(self._domain.storage_devices):
            self.debug(f'final domain disk #{i} size', str(device.size))

        # Is this a CoreOS?
        self._domain.coreos = self.is_coreos

        self._apply_hw_arch(self._domain, self.is_kvm, self.is_legacy_os)

        mac_address = testcloud.util.generate_mac_address()
        if f"qemu:///{self.connection}" == "qemu:///system":
            self._domain.network_configuration = SystemNetworkConfiguration(
                mac_address=mac_address
            )
        elif f"qemu:///{self.connection}" == "qemu:///session":
            device_type = "virtio-net-pci" if not self.is_legacy_os else "e1000"
            with GuestTestcloud._testcloud_lock:
                port = testcloud.util.spawn_instance_port_file(self.instance_name)
            self._domain.network_configuration = UserNetworkConfiguration(
                mac_address=mac_address, port=port, device_type=device_type
            )
        else:
            raise tmt.utils.ProvisionError("Only system, or session connection is supported.")

        if not self._domain.coreos:
            seed_disk = RawStorageDevice(self._domain.seed_path)
            self._domain.storage_devices.append(seed_disk)

        self._instance = testcloud.instance.Instance(
            hostname=hostname,
            image=self._image,
            connection=f"qemu:///{self.connection}",
            domain_configuration=self._domain,
            workarounds=self._workarounds,
        )

        self.verbose('name', self.instance_name, 'green')

        # Decide if we want to multiply timeouts when emulating an architecture
        time_coeff = NON_KVM_TIMEOUT_COEF if not self.is_kvm else 1

        # Prepare ssh key
        # TODO: Maybe... some better way to do this?
        public_key = self.prepare_ssh_key(SSH_KEYGEN_TYPE)
        if self._domain.coreos:
            self._instance.coreos = True
            # prepare_ssh_key() writes key directly to COREOS_DATA
            self._instance.ssh_path = []
            data_tpl = Template(COREOS_DATA).safe_substitute(
                user_name=self.user, public_key=public_key
            )
        else:
            data_tpl = Template(USER_DATA).safe_substitute(
                user_name=self.user, public_key=public_key
            )

        # Boot the virtual machine
        self.info('progress', 'booting...', 'cyan')
        assert libvirt is not None

        try:
            self._instance.prepare(data_tpl=data_tpl)
            self._instance.spawn_vm()
            self._instance.start(BOOT_TIMEOUT * time_coeff)
        except (testcloud.exceptions.TestcloudInstanceError, libvirt.libvirtError) as error:
            raise ProvisionError(f'Failed to boot testcloud instance ({error}).')
        self.primary_address = self.topology_address = self._instance.get_ip()
        self.port = int(self._instance.get_instance_port())
        self.verbose('primary address', self.primary_address, 'green')
        self.verbose('topology address', self.topology_address, 'green')
        self.verbose('port', self.port, 'green')
        self._instance.create_ip_file(self.primary_address)

        # Wait a bit until the box is up
        if not self.reconnect(
            Waiting(Deadline.from_seconds(CONNECT_TIMEOUT * time_coeff), tick=1)
        ):
            raise ProvisionError(f"Failed to connect in {CONNECT_TIMEOUT * time_coeff}s.")

    def stop(self) -> None:
        """
        Stop provisioned guest
        """

        super().stop()
        # Stop only if the instance successfully booted
        if self._instance and self.primary_address:
            self.debug(f"Stopping testcloud instance '{self.instance_name}'.")
            assert testcloud is not None
            try:
                self._instance.stop()
            except testcloud.exceptions.TestcloudInstanceError as error:
                raise tmt.utils.ProvisionError(f"Failed to stop testcloud instance: {error}")

            self.info('guest', 'stopped', 'green')

    def remove(self) -> None:
        """
        Remove the guest (disk cleanup)
        """

        if self._instance:
            self.debug(f"Removing testcloud instance '{self.instance_name}'.")
            try:
                self._instance.remove(autostop=True)
            except FileNotFoundError as error:
                raise tmt.utils.ProvisionError(f"Failed to remove testcloud instance: {error}")

            self.info('guest', 'removed', 'green')

    def reboot(
        self,
        hard: bool = False,
        command: Optional[Union[Command, ShellScript]] = None,
        waiting: Optional[Waiting] = None,
    ) -> bool:
        """
        Reboot the guest, and wait for the guest to recover.

        .. note::

           Custom reboot command can be used only in combination with a
           soft reboot. If both ``hard`` and ``command`` are set, a hard
           reboot will be requested, and ``command`` will be ignored.

        :param hard: if set, force the reboot. This may result in a loss
            of data. The default of ``False`` will attempt a graceful
            reboot.
        :param command: a command to run on the guest to trigger the
            reboot. If ``hard`` is also set, ``command`` is ignored.
        :param timeout: amount of time in which the guest must become available
            again.
        :param tick: how many seconds to wait between two consecutive attempts
            of contacting the guest.
        :param tick_increase: a multiplier applied to ``tick`` after every
            attempt.
        :returns: ``True`` if the reboot succeeded, ``False`` otherwise.
        """

        waiting = waiting or tmt.steps.provision.default_reboot_waiting()

        if hard:
            if self._instance is None:
                raise tmt.utils.ProvisionError("No instance initialized.")

            self.debug("Hard reboot using the testcloud API.")

            # ignore[union-attr]: mypy still considers `self._instance` as possibly
            # being `None`, missing the explicit check above.
            return self.perform_reboot(
                lambda: self._instance.reboot(soft=False),  # type: ignore[union-attr]
                waiting,
                fetch_boot_time=False,
            )

        if command:
            return super().reboot(
                hard=False,
                command=command,
                waiting=waiting,
            )

        if self._instance is None:
            raise tmt.utils.ProvisionError("No instance initialized.")

        # ignore[union-attr]: mypy still considers `self._instance` as possibly
        # being `None`, missing the explicit check above.
        return self.perform_reboot(
            lambda: self._instance.reboot(soft=True),  # type: ignore[union-attr]
            waiting,
        )


@tmt.steps.provides_method(
    'virtual.testcloud',
    installation_hint="""
        Make sure ``testcloud`` and ``libvirt`` packages are installed and configured, they are
        required for VM-backed guests provided by ``provision/virtual.testcloud`` plugin.

        * Users who installed tmt from system repositories should install ``tmt+provision-virtual``
          package.
        * Users who installed tmt from PyPI should also install ``tmt+provision-virtual`` package,
          as it will install required system dependencies. After doing so, they should install
          ``tmt[provision-virtual]`` extra.
    """,
)
class ProvisionTestcloud(tmt.steps.provision.ProvisionPlugin[ProvisionTestcloudData]):
    """
    Local virtual machine using ``testcloud`` library.
    Testcloud takes care of downloading an image and
    making necessary changes to it for optimal experience
    (such as disabling ``UseDNS`` and ``GSSAPI`` for SSH).

    Minimal config which uses the latest Fedora image:

    .. code-block:: yaml

        provision:
            how: virtual

    Here's a full config example:

    .. code-block:: yaml

        # Provision a virtual machine from a specific QCOW2 file,
        # using specific memory and disk settings, using the fedora user,
        # and using sudo to run scripts.
        provision:
            how: virtual
            image: https://mirror.vpsnet.com/fedora/linux/releases/41/Cloud/x86_64/images/Fedora-Cloud-Base-Generic-41-1.4.x86_64.qcow2
            user: fedora
            become: true
            # in MB
            memory: 2048
            # in GB
            disk: 30

    As the image use ``fedora`` for the latest released Fedora compose,
    ``fedora-rawhide`` for the latest Rawhide compose, short aliases such as
    ``fedora-32``, ``f-32`` or ``f32`` for specific release or a full url to
    the qcow2 image for example from https://kojipkgs.fedoraproject.org/compose/.

    Short names are also provided for ``centos``, ``centos-stream``, ``alma``,
    ``rocky``, ``oracle``, ``debian`` and ``ubuntu`` (e.g. ``centos-8`` or ``c8``).

    .. note::

        The non-rpm distros are not fully supported yet in tmt as
        the package installation is performed solely using ``dnf``/``yum``
        and ``rpm``.
        But you should be able the login to the provisioned guest and start
        experimenting. Full support is coming in the future :)

    Supported Fedora CoreOS images are:

    * ``fedora-coreos``
    * ``fedora-coreos-stable``
    * ``fedora-coreos-testing``
    * ``fedora-coreos-next``

    Use the full path for images stored on local disk, for example:

    .. code-block:: shell

        /var/tmp/images/Fedora-Cloud-Base-31-1.9.x86_64.qcow2

    In addition to the qcow2 format, Vagrant boxes can be used as well,
    testcloud will take care of unpacking the image for you.

    To trigger hard reboot of a guest, plugin uses testcloud API. It is
    also used to trigger soft reboot unless a custom reboot command was
    specified via ``tmt-reboot -c ...``.
    """

    _data_class = ProvisionTestcloudData
    _guest_class = GuestTestcloud

    _thread_safe = True

    # Guest instance
    _guest = None

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Provision the testcloud instance
        """

        super().go(logger=logger)

        if self.data.list_local_images:
            self._print_local_images()
            # Clean up the run workdir and exit
            if self.step.plan.my_run:
                self.step.plan.my_run._workdir_cleanup()
            raise SystemExit(0)

        # Give info about provided data
        data = TestcloudGuestData.from_plugin(self)

        # Once plan schema is enforced this won't be necessary
        # click enforces int for cmdline and schema validation
        # will make sure 'int' gets from plan data.
        # Another key is 'port' however that is not exposed to the cli
        for int_key in ["port"]:
            value = getattr(data, int_key)
            if value is not None:
                try:
                    setattr(data, int_key, int(value))
                except ValueError as exc:
                    raise tmt.utils.NormalizationError(
                        f'{self.name}:{int_key}', value, 'an integer'
                    ) from exc

        data.show(verbose=self.verbosity_level, logger=self._logger)

        if data.hardware and data.hardware.constraint:
            data.hardware.report_support(check=_report_hw_requirement_support, logger=self._logger)

            for line in data.hardware.format_variants():
                self._logger.debug('hardware', line, level=4)

            if data.memory is not None and data.hardware.constraint.uses_constraint(
                'memory', self._logger
            ):
                self._logger.warning(
                    "Hardware requirement 'memory' is specified in 'hardware' key,"
                    " it will be overruled by 'memory' key."
                )

            if data.disk is not None and data.hardware.constraint.uses_constraint(
                'disk.size', self._logger
            ):
                self._logger.warning(
                    "Hardware requirement 'disk.size' is specified in 'hardware' key,"
                    " it will be overruled by 'disk' key."
                )

        # Create a new GuestTestcloud instance and start it
        self._guest = GuestTestcloud(
            logger=self._logger,
            data=data,
            name=self.name,
            parent=self.step,
        )
        self._guest.start()
        self._guest.setup()

    def _print_local_images(self) -> None:
        """
        Print images which are already cached
        """

        self.info("Locally available images")
        for filename in sorted(TESTCLOUD_IMAGES.glob('*.qcow2')):
            self.info(filename.name, shift=1, color='yellow')
            click.echo(f"{TESTCLOUD_IMAGES / filename}")

    @classmethod
    def clean_images(cls, clean: 'tmt.base.Clean', dry: bool, workdir_root: Path) -> bool:
        """
        Remove the testcloud images
        """

        testcloud_images = workdir_root / 'testcloud/images'
        clean.info('testcloud', shift=1, color='green')
        if not testcloud_images.exists():
            clean.warn(f"Directory '{testcloud_images}' does not exist.", shift=2)
            return True
        successful = True
        for image in testcloud_images.iterdir():
            if dry:
                clean.verbose(f"Would remove '{image}'.", shift=2)
            else:
                clean.verbose(f"Removing '{image}'.", shift=2)
                try:
                    image.unlink()
                except OSError:
                    clean.fail(f"Failed to remove '{image}'.", shift=2)
                    successful = False
        return successful



import urllib.parse

from tmt.utils import Path
from tmt.utils.git import DistGitHandler

MOCK_SOURCES_FILENAME = 'mock_sources'
SERVER_PORT = 9000


class TestDistGit(DistGitHandler):
    """
    Test handler

    each line in MOCK_SOURCES_FILENAME contains
    filename to serve and optional (after single space a custom filename)
    """

    usage_name = "TESTING"
    server = f"http://localhost:{SERVER_PORT}"

    def url_and_name(self, cwd=None):
        cwd = cwd or Path.cwd()
        data = (cwd / MOCK_SOURCES_FILENAME).read_text()
        for line in data.splitlines():
            split = line.split(' ', maxsplit=2)
            url = split[0]
            try:
                src_name = split[1]
            except IndexError:
                src_name = url

            yield (urllib.parse.urljoin(self.server, url), src_name)



import tmt.steps
import tmt.steps.finish
import tmt.steps.prepare.ansible
from tmt.steps.prepare.ansible import PrepareAnsible


@tmt.steps.provides_method('ansible')
class FinishAnsible(
    tmt.steps.finish.FinishPlugin[tmt.steps.finish.FinishStepData], PrepareAnsible
):
    """
    Perform finishing tasks using ansible.

    One or more playbooks can be provided as a list under the ``playbook``
    attribute.  Each of them will be applied using ``ansible-playbook`` in
    the given order. The path must be relative to the metadata tree root.

    Remote playbooks can be referenced as well as the local ones,
    and both kinds can be used at the same time.

    .. warning::

       The plugin may be a subject of various limitations, imposed by
       Ansible itself:

       * Ansible 2.17+ no longer supports Python 3.6 and older. Guests
         where Python 3.7+ is not available cannot be prepared with the
         ``ansible`` plugin. This has been observed when Fedora Rawhide
         runner is used with CentOS 7 or CentOS Stream 8 guests. Possible
         workarounds: downgrade Ansible tmt uses, or install Python 3.7+
         before using ``ansible`` plugin from an alternative repository
         or local build.

    Single playbook config:

    .. code-block:: yaml

        finish:
            how: ansible
            playbook: ansible/packages.yml

    Multiple playbooks config:

    .. code-block:: yaml

        finish:
            how: ansible
            playbook:
              - playbooks/common.yml
              - playbooks/os/rhel9.yml
              - https://foo.bar/rhel9-final-touches.yml

    The playbook path should be relative to the metadata tree root. Use
    the :ref:`/spec/core/order` attribute to select in which order
    finishing tasks should happen if there are multiple configs. Default
    order is ``50``.
    """

    # We are reusing "prepare" step for "finish",
    # and they both have different expectations
    _data_class = tmt.steps.prepare.ansible.PrepareAnsibleData

    # FIXME: ignore[assignment]: https://github.com/teemtee/tmt/issues/1540
    # Also, assigning class methods seems to cause trouble to mypy
    # See https://github.com/python/mypy/issues/6700
    base_command = tmt.steps.finish.FinishPlugin.base_command  # type: ignore[assignment]

    # `FinishPlugin` plugin would win the inheritance battle and provide
    # its no-op `go()`. Force the one from `PrepareAnsible`.
    go = PrepareAnsible.go
